# -*- coding: utf-8 -*-

"""This test module tests classes, methods and functions of the core/parameters
module.
"""

import numpy as np
import os.path
import sys
import unittest

from skyllh.core.binning import BinningDefinition
from skyllh.core.model import Model
from skyllh.core.parameters import (
    HypoParameterDefinition,
    Parameter,
    ParameterSet,
    ParameterSetArray,
    ParameterGrid,
    ParameterGridSet,
    SingleModelParameterMapper,
    MultiModelParameterMapper
)
from skyllh.core.py import const
from skyllh.core.random import RandomStateService

sys.path.append(os.path.join(os.path.split(__file__)[0], '..'))
from utils import isAlmostEqual

GAMMA_GRID = [
    1. ,  1.1,  1.2,  1.3,  1.4,  1.5,  1.6,  1.7,  1.8,  1.9,  2.
]

ECUT_GRID = [
    9., 9.1
]


class Parameter_TestCase(unittest.TestCase):
    """This test case tests the Parameter class.
    """
    def setUp(self):
        self.fixed_param_initial = 2.37
        self.floating_param_initial = 7.32
        self.floating_param_valmin = 7.1
        self.floating_param_valmax = 8
        self.floating_param_grid = np.array([
            7.1, 7.2, 7.3, 7.4, 7.5, 7.6, 7.7, 7.8, 7.9, 8.])

        self.fixed_param = Parameter(
            'fixed_param', self.fixed_param_initial)
        self.floating_param = Parameter(
            'floating_param', self.floating_param_initial,
            valmin=self.floating_param_valmin,
            valmax=self.floating_param_valmax)

    def test__eq__(self):
        fixed_param = Parameter(
            'fixed_param', self.fixed_param_initial)
        floating_param = Parameter(
            'floating_param', self.floating_param_initial,
            valmin=self.floating_param_valmin,
            valmax=self.floating_param_valmax)

        self.assertTrue(self.fixed_param == fixed_param)
        self.assertTrue(self.floating_param == floating_param)

        # Change the parameter name.
        fixed_param = Parameter(
            'fixed_param1', self.fixed_param_initial)
        self.assertFalse(self.fixed_param == fixed_param)

        # Change the initial value.
        fixed_param = Parameter(
            'fixed_param', self.fixed_param_initial+1)
        self.assertFalse(self.fixed_param == fixed_param)

        floating_param = Parameter(
            'floating_param', self.floating_param_initial+0.2,
            valmin=self.floating_param_valmin,
            valmax=self.floating_param_valmax)
        self.assertFalse(self.floating_param == floating_param)

        # Change the valmin.
        floating_param = Parameter(
            'floating_param', self.floating_param_initial,
            valmin=self.floating_param_valmin-1,
            valmax=self.floating_param_valmax)
        self.assertFalse(self.floating_param == floating_param)

        # Change the valmax.
        floating_param = Parameter(
            'floating_param', self.floating_param_initial,
            valmin=self.floating_param_valmin,
            valmax=self.floating_param_valmax+1)
        self.assertFalse(self.floating_param == floating_param)

    def test_name(self):
        self.assertEqual(self.fixed_param.name, 'fixed_param')
        self.assertEqual(self.floating_param.name, 'floating_param')

    def test_initial(self):
        self.assertTrue(isAlmostEqual(
            self.fixed_param.initial, self.fixed_param_initial))
        self.assertTrue(isAlmostEqual(
            self.floating_param.initial, self.floating_param_initial))

    def test_isfixed(self):
        self.assertTrue(self.fixed_param.isfixed)
        self.assertFalse(self.floating_param.isfixed)

    def test_valmin(self):
        self.assertEqual(self.fixed_param.valmin, None)
        self.assertTrue(isAlmostEqual(
            self.floating_param.valmin, self.floating_param_valmin))

    def test_valmax(self):
        self.assertEqual(self.fixed_param.valmax, None)
        self.assertTrue(isAlmostEqual(
            self.floating_param.valmax, self.floating_param_valmax))

    def test_value(self):
        self.assertTrue(isAlmostEqual(
            self.fixed_param.value, self.fixed_param_initial))
        self.assertTrue(isAlmostEqual(
            self.floating_param.value, self.floating_param_initial))

        # Try to change the value of a fixed parameter.
        with self.assertRaises(ValueError):
            self.fixed_param.value = self.floating_param_initial

        # Try to set the value of a floating parameter to a value outside its
        # value range.
        with self.assertRaises(ValueError):
            self.floating_param.value = self.fixed_param_initial

    def test_str(self):
        try:
            str(self.fixed_param)
            str(self.floating_param)
        except:
            self.fail('The __str__ method raised an exception!')

    def test_as_linear_grid(self):
        grid_delta = 0.1
        with self.assertRaises(ValueError):
            self.fixed_param.as_linear_grid(grid_delta)

        param_grid = self.floating_param.as_linear_grid(grid_delta)
        self.assertTrue(np.all(isAlmostEqual(
            param_grid.grid, self.floating_param_grid)))

    def test_change_fixed_value(self):
        with self.assertRaises(ValueError):
            self.floating_param.change_fixed_value(self.fixed_param_initial)

        self.fixed_param.change_fixed_value(self.floating_param_initial)
        self.assertTrue(isAlmostEqual(
            self.fixed_param.initial, self.floating_param_initial))
        self.assertTrue(isAlmostEqual(
            self.fixed_param.value, self.floating_param_initial))

    def test_make_fixed(self):
        self.floating_param.make_fixed(self.fixed_param_initial)
        self.assertTrue(isAlmostEqual(
            self.floating_param.initial, self.fixed_param_initial))
        self.assertTrue(isAlmostEqual(
            self.floating_param.value, self.fixed_param_initial))

    def test_make_floating(self):
        with self.assertRaises(ValueError):
            self.fixed_param.make_floating()
        with self.assertRaises(ValueError):
            self.fixed_param.make_floating(valmin=self.floating_param_valmin)

        # The current value of fixed_param is outside the valmin and valmax
        # range of the floating_param. This should raise an exception when no
        # new initial value is specified.
        with self.assertRaises(ValueError):
            self.fixed_param.make_floating(
                valmin=self.floating_param_valmin,
                valmax=self.floating_param_valmax)

        self.fixed_param.make_floating(
            initial=self.floating_param_initial,
            valmin=self.floating_param_valmin,
            valmax=self.floating_param_valmax)
        self.assertTrue(isAlmostEqual(
            self.fixed_param.initial, self.floating_param_initial))
        self.assertTrue(isAlmostEqual(
            self.fixed_param.value, self.floating_param_initial))
        self.assertTrue(isAlmostEqual(
            self.fixed_param.valmin, self.floating_param_valmin))
        self.assertTrue(isAlmostEqual(
            self.fixed_param.valmax, self.floating_param_valmax))


class ParameterSet_TestCase(unittest.TestCase):
    """This test case tests the ParameterSet class.
    """
    def setUp(self):
        self.fixed_param = Parameter('p0', 2.3)
        self.floating_param = Parameter('p1', 1.1, valmin=0.5, valmax=1.6)
        self.paramset = ParameterSet((self.fixed_param, self.floating_param))

    def test_union(self):
        p0 = Parameter('p0', 2.3)
        p1 = Parameter('p1', 1.1, valmin=0.5, valmax=1.6)
        p2 = Parameter('p2', 3.2, valmin=2.3, valmax=4.7)
        paramset0 = ParameterSet((p0,p2))
        paramset1 = ParameterSet((p1,p2))
        paramset_union = ParameterSet.union(paramset0, paramset1)
        params = paramset_union.params
        self.assertEqual(len(params), 3)
        self.assertEqual(params[0], p0)
        self.assertEqual(params[1], p2)
        self.assertEqual(params[2], p1)

    def test_params(self, paramset=None):
        if(paramset is None):
            paramset = self.paramset
        params = self.paramset.params
        self.assertEqual(len(params), 2)
        self.assertEqual(params[0], self.fixed_param)
        self.assertEqual(params[1], self.floating_param)

    def test_fixed_params(self):
        params = self.paramset.fixed_params
        self.assertEqual(len(params), 1)
        self.assertEqual(params[0], self.fixed_param)

    def test_fixed_params_mask(self):
        mask = self.paramset.fixed_params_mask
        self.assertTrue(len(mask), 2)
        self.assertEqual(mask[0], True)
        self.assertEqual(mask[1], False)

    def test_floating_params(self):
        params = self.paramset.floating_params
        self.assertEqual(len(params), 1)
        self.assertEqual(params[0], self.floating_param)

    def test_floating_params_mask(self):
        mask = self.paramset.floating_params_mask
        self.assertTrue(len(mask), 2)
        self.assertEqual(mask[0], False)
        self.assertEqual(mask[1], True)

    def test_n_params(self):
        self.assertEqual(self.paramset.n_params, 2)

    def test_n_fixed_params(self):
        self.assertEqual(self.paramset.n_fixed_params, 1)

    def test_n_floating_params(self):
        self.assertEqual(self.paramset.n_floating_params, 1)

    def test_fixed_param_name_list(self):
        names = self.paramset.fixed_param_name_list
        self.assertEqual(len(names), 1)
        self.assertEqual(names, ['p0'])

    def test_floating_param_name_list(self):
        names = self.paramset.floating_param_name_list
        self.assertEqual(len(names), 1)
        self.assertEqual(names, ['p1'])

    def test_fixed_param_values(self):
        values = self.paramset.fixed_param_values
        self.assertTrue(isAlmostEqual(values, [2.3]))

    def test_floating_param_initials(self):
        initials = self.paramset.floating_param_initials
        self.assertTrue(isAlmostEqual(initials, [1.1]))

    def test_floating_param_bounds(self):
        bounds = self.paramset.floating_param_bounds
        self.assertTrue(isAlmostEqual(bounds[0], [0.5, 1.6]))

    def test_len(self):
        self.assertEqual(len(self.paramset), 2)

    def test_iter(self):
        for (i,param) in enumerate(self.paramset):
            self.assertEqual(param.name, 'p%d'%(i))

    def test_str(self):
        try:
            str(self.paramset)
        except:
            self.fail('The __str__ method raised exception!')

    def test_get_fixed_pidx(self):
        self.assertEqual(self.paramset.get_fixed_pidx('p0'), 0)

        with self.assertRaises(KeyError):
            self.paramset.get_fixed_pidx('p1')

    def test_get_floating_pidx(self):
        self.assertEqual(self.paramset.get_floating_pidx('p1'), 0)

        with self.assertRaises(KeyError):
            self.paramset.get_floating_pidx('p0')

    def test_has_fixed_param(self):
        self.assertTrue(self.paramset.has_fixed_param('p0'))
        self.assertFalse(self.paramset.has_fixed_param('p1'))

    def test_has_floating_param(self):
        self.assertTrue(self.paramset.has_floating_param('p1'))
        self.assertFalse(self.paramset.has_floating_param('p0'))

    def test_make_params_fixed(self):
        # Already fixed parameters cannot be fixed.
        with self.assertRaises(ValueError):
            self.paramset.make_params_fixed({'p0': 42})

        # Fix the floating parameter outside its current range.
        self.paramset.make_params_fixed({'p1': 0.4})
        self.assertTrue(self.paramset.has_fixed_param('p1'))
        self.assertEqual(self.paramset.n_fixed_params, 2)
        self.assertEqual(self.paramset.n_floating_params, 0)
        self.test_params()
        values = self.paramset.fixed_param_values
        self.assertTrue(isAlmostEqual(values, [2.3, 0.4]))
        self.assertEqual(self.paramset.params[1].valmin, None)
        self.assertEqual(self.paramset.params[1].valmax, None)

        self.setUp()

        # Fix the floating parameter to its current value.
        self.paramset.make_params_fixed({'p1': None})
        self.assertTrue(self.paramset.has_fixed_param('p1'))
        self.assertEqual(self.paramset.n_fixed_params, 2)
        self.assertEqual(self.paramset.n_floating_params, 0)
        self.test_params()
        values = self.paramset.fixed_param_values
        self.assertTrue(isAlmostEqual(values, [2.3, 1.1]))
        self.assertTrue(isAlmostEqual(self.paramset.params[1].valmin, 0.5))
        self.assertTrue(isAlmostEqual(self.paramset.params[1].valmax, 1.6))

    def test_make_params_floating(self):
        # Already floating parameters cannot be made floating.
        with self.assertRaises(ValueError):
            self.paramset.make_params_floating({'p1': None})

        # Make the fixed parameter floating.
        with self.assertRaises(ValueError):
            self.paramset.make_params_floating({'p0': None})
        with self.assertRaises(ValueError):
            self.paramset.make_params_floating({'p0': 1.2})
        self.paramset.make_params_floating({'p0': (1.2, 1.0, 1.3)})
        self.assertTrue(self.paramset.has_floating_param('p0'))
        self.assertTrue(isAlmostEqual(self.paramset.params[0].initial, 1.2))
        self.assertTrue(isAlmostEqual(self.paramset.params[0].valmin, 1.0))
        self.assertTrue(isAlmostEqual(self.paramset.params[0].valmax, 1.3))

    def test_update_fixed_param_value_cache(self):
        self.assertAlmostEqual(self.paramset.params[0].value, 2.3)
        self.fixed_param.change_fixed_value(3.1)
        self.assertAlmostEqual(self.paramset.params[0].value, 3.1)
        self.assertTrue(isAlmostEqual(self.paramset.fixed_param_values, [2.3]))
        self.paramset.update_fixed_param_value_cache()
        self.assertTrue(isAlmostEqual(self.paramset.fixed_param_values, [3.1]))

    def test_copy(self):
        new_paramset = self.paramset.copy()
        self.assertFalse(new_paramset == self.paramset)
        self.test_params(new_paramset)

    def test_add_param(self):
        with self.assertRaises(TypeError):
            self.paramset.add_param('p2')
        with self.assertRaises(KeyError):
            param = Parameter('p0', 42.)
            self.paramset.add_param(param)

        # Add parameter at front.
        param = Parameter('p2', 42.)
        self.paramset.add_param(param, atfront=True)
        self.assertEqual(self.paramset.params[0], param)
        self.assertEqual(self.paramset.params[1], self.fixed_param)
        self.assertEqual(self.paramset.params[2], self.floating_param)

        self.setUp()

        # Add parameter at end.
        param = Parameter('p2', 42.)
        self.paramset.add_param(param)
        self.assertEqual(self.paramset.params[0], self.fixed_param)
        self.assertEqual(self.paramset.params[1], self.floating_param)
        self.assertEqual(self.paramset.params[2], param)

    def test_has_param(self):
        self.assertTrue(self.paramset.has_param(self.fixed_param))
        self.assertTrue(self.paramset.has_param(self.floating_param))
        self.assertFalse(self.paramset.has_param(Parameter('p', 0.0)))

    def test_floating_param_values_to_dict(self):
        param_dict = self.paramset.floating_param_values_to_dict(np.array([1.3]))
        self.assertAlmostEqual(param_dict['p0'], 2.3)
        self.assertAlmostEqual(param_dict['p1'], 1.3)


class ParameterSetArray_TestCase(unittest.TestCase):
    """This test case tests the ParameterSetArray class.
    """
    def setUp(self):
        self.fixed_param0 = Parameter(
            'fixed_param0', 2.3)
        self.fixed_param1 = Parameter(
            'fixed_param1', 0.1)
        self.floating_param0 = Parameter(
            'floating_param0', 1.1, valmin=0.5, valmax=1.6)
        self.floating_param1 = Parameter(
            'floating_param1', 13.5, valmin=10.5, valmax=15)
        self.paramset0 = ParameterSet((self.fixed_param0, self.floating_param0))
        self.paramset1 = ParameterSet((self.floating_param1, self.fixed_param1))

        self.paramsetarr = ParameterSetArray(
            (const(self.paramset0), const(self.paramset1)))

    def test__init__(self):
        paramset0 = ParameterSet((self.fixed_param0, self.floating_param0))
        paramset1 = ParameterSet((self.floating_param1, self.fixed_param1))
        with self.assertRaises(TypeError):
            paramsetarr = ParameterSetArray(
                (paramset0, paramset1))

    def test_paramset_list(self):
        paramset_list = self.paramsetarr.paramset_list
        self.assertEqual(len(paramset_list), 2)
        self.assertEqual(paramset_list[0], self.paramset0)
        self.assertEqual(paramset_list[1], self.paramset1)

    def test_n_params(self):
        self.assertEqual(self.paramsetarr.n_params, 4)

    def test_n_fixed_params(self):
        self.assertEqual(self.paramsetarr.n_fixed_params, 2)

    def test_n_floating_params(self):
        self.assertEqual(self.paramsetarr.n_floating_params, 2)

    def test_floating_param_initials(self):
        initials = self.paramsetarr.floating_param_initials
        self.assertEqual(initials.shape, (2,))
        self.assertAlmostEqual(initials[0], 1.1)
        self.assertAlmostEqual(initials[1], 13.5)

    def test_floating_param_bounds(self):
        bounds = self.paramsetarr.floating_param_bounds
        self.assertEqual(bounds.shape, (2,2))
        np.testing.assert_almost_equal(bounds[0], (0.5, 1.6))
        np.testing.assert_almost_equal(bounds[1], (10.5, 15))

    def test__str__(self):
        try:
            str(self.paramsetarr)
        except:
            self.fail('The __str__ method raised exception!')

    def test_generate_random_initials(self):
        rss_ref = RandomStateService(42)
        rn = rss_ref.random.uniform(size=2)

        rss = RandomStateService(42)
        initials = self.paramsetarr.generate_random_initials(rss)
        np.testing.assert_almost_equal(initials,
            (0.5+rn[0]*(1.6-0.5),
            10.5+rn[1]*(15-10.5)))

    def test_split_floating_param_values(self):
        fl_param_values = np.array([0.9, 14.2])
        fl_param_values_list = self.paramsetarr.split_floating_param_values(
            fl_param_values)
        self.assertEqual(len(fl_param_values_list), 2)
        self.assertEqual(len(fl_param_values_list[0]), 1)
        self.assertEqual(len(fl_param_values_list[1]), 1)
        np.testing.assert_almost_equal(fl_param_values_list[0], [0.9])
        np.testing.assert_almost_equal(fl_param_values_list[1], [14.2])


class ParameterGrid_TestCase(unittest.TestCase):
    """This test case tests the ParameterGrid class.
    """
    def setUp(self):
        self.paramgrid_gamma1 = ParameterGrid('gamma1', [ 1.5, 2., 2.5, 3., 3.5])
        self.paramgrid_gamma2 = ParameterGrid('gamma2', GAMMA_GRID)
        self.paramgrid_gamma3 = ParameterGrid('gamma3', [ 1.05, 1.15, 1.25, 1.35])

    def test_from_BinningDefinition(self):
        binning = BinningDefinition(name='gamma', binedges=GAMMA_GRID)
        param_grid = ParameterGrid.from_BinningDefinition(binning)

        self.assertEqual(param_grid.name, binning.name)
        self.assertTrue(isAlmostEqual(param_grid.grid, GAMMA_GRID))

    def test_delta(self):
        self.assertTrue(isAlmostEqual(self.paramgrid_gamma1.delta, 0.5))
        self.assertTrue(isAlmostEqual(self.paramgrid_gamma2.delta, 0.1))
        self.assertTrue(isAlmostEqual(self.paramgrid_gamma3.delta, 0.1))

    def test_decimals(self):
        self.assertTrue(self.paramgrid_gamma1.decimals >= 1)
        self.assertTrue(self.paramgrid_gamma2.decimals >= 1)
        self.assertTrue(self.paramgrid_gamma3.decimals >= 2)

    def test_round_to_nearest_grid_point(self):
        # Test values outside the grid range.
        x = 1.49999999999
        with self.assertRaises(ValueError):
            gp = self.paramgrid_gamma1.round_to_nearest_grid_point(x)
        x = 3.50000000001
        with self.assertRaises(ValueError):
            gp = self.paramgrid_gamma1.round_to_nearest_grid_point(x)

        # Test a value between two grid points.
        x = [2.1, 2.4, 2.2, 2.3]
        gp = self.paramgrid_gamma1.round_to_nearest_grid_point(x)
        np.testing.assert_almost_equal(gp, [2.0, 2.5, 2., 2.5])

        x = [1.051, 1.14]
        gp = self.paramgrid_gamma3.round_to_nearest_grid_point(x)
        self.assertTrue(isAlmostEqual(gp, [1.05, 1.15]))

        # Test a value on a grid point.
        x = [1.05, 1.35]
        gp = self.paramgrid_gamma3.round_to_nearest_grid_point(x)
        self.assertTrue(isAlmostEqual(gp, [1.05, 1.35]))

    def test_round_to_lower_grid_point(self):
        # Test a value between two grid points.
        x = 2.4
        gp = self.paramgrid_gamma1.round_to_lower_grid_point(x)
        self.assertTrue(isAlmostEqual(gp, 2.))

        # Test a value at a grid point.
        x = 2.
        gp = self.paramgrid_gamma1.round_to_lower_grid_point(x)
        self.assertTrue(isAlmostEqual(gp, 2.))

        x = 1.6
        gp = self.paramgrid_gamma2.round_to_lower_grid_point(x)
        self.assertTrue(isAlmostEqual(gp, 1.6))

        x = [1.05, 1.15, 1.25, 1.35]
        gp = self.paramgrid_gamma3.round_to_lower_grid_point(x)
        np.testing.assert_almost_equal(gp, [1.05, 1.15, 1.25, 1.35])

    def test_round_to_upper_grid_point(self):
        # Test a value between two grid points.
        x = 2.4
        gp = self.paramgrid_gamma1.round_to_upper_grid_point(x)
        self.assertTrue(isAlmostEqual(gp, 2.5))

        # Test a value at a grid point.
        x = 2.
        gp = self.paramgrid_gamma1.round_to_upper_grid_point(x)
        self.assertTrue(isAlmostEqual(gp, 2.5))

        x = 1.6
        gp = self.paramgrid_gamma2.round_to_upper_grid_point(x)
        self.assertTrue(isAlmostEqual(gp, 1.7))

        x = [1.05, 1.15, 1.25, 1.35]
        gp = self.paramgrid_gamma3.round_to_upper_grid_point(x)
        np.testing.assert_almost_equal(gp, [1.15, 1.25, 1.35, 1.45])


class ParameterGridSet_TestCase(unittest.TestCase):
    """This test case tests the ParameterGridSet class.
    """
    def setUp(self):
        """Setups this test case.
        """
        self.paramgrid_gamma = ParameterGrid('gamma', GAMMA_GRID)
        self.paramgrid_Ecut = ParameterGrid('Ecut', ECUT_GRID)

        self.paramgridset = ParameterGridSet(
            (self.paramgrid_gamma, self.paramgrid_Ecut))

    def test_ndim(self):
        self.assertEqual(
            self.paramgridset.ndim, 2)

    def test_parameter_names(self):
        self.assertEqual(
            self.paramgridset.parameter_names, ['gamma', 'Ecut'])

    def test_parameter_permutation_dict_list(self):
        perm_dict_list = self.paramgridset.parameter_permutation_dict_list

        self.assertTrue(isAlmostEqual(
            [ d['gamma'] for d in perm_dict_list ],
            np.repeat(np.array(GAMMA_GRID), len(ECUT_GRID))
        ))
        self.assertTrue(isAlmostEqual(
            [ d['Ecut'] for d in perm_dict_list ],
            list(ECUT_GRID)*len(GAMMA_GRID)
        ))

    def test_index(self):
        self.assertEqual(
            self.paramgridset.index(self.paramgrid_gamma), 0)

        self.assertEqual(
            self.paramgridset.index(self.paramgrid_Ecut), 1)

    def test_index_by_name(self):
        self.assertEqual(
            self.paramgridset.index_by_name('gamma'), 0)

        self.assertEqual(
            self.paramgridset.index_by_name('Ecut'), 1)

    def test_pop_and_add(self):
        paramgrid_gamma = self.paramgridset.pop('gamma')
        self.assertEqual(paramgrid_gamma.name, 'gamma')

        paramgrid_Ecut = self.paramgridset.pop()
        self.assertEqual(paramgrid_Ecut.name, 'Ecut')

        self.paramgridset.add(paramgrid_gamma)
        self.paramgridset.add(paramgrid_Ecut)

        # The altered ParameterGridSet instance should be the same as the
        # initial ParameterGridSet instance. So just run all the tests on that
        # altered one.
        self.test_ndim()
        self.test_parameter_names()
        self.test_parameter_permutation_dict_list()
        self.test_index()
        self.test_index_by_name()


class SingleModelParameterMapperTestCase(unittest.TestCase):
    def setUp(self):
        self.fixed_param = Parameter('p1', 42)
        self.floating_param = Parameter('p2', 4, 1, 6)
        self.model = Model('the_src')
        self.mpm = SingleModelParameterMapper(
            'signal', self.model)

    def test_name(self):
        self.assertEqual(self.mpm.name, 'signal')

    def test_models(self):
        self.assertEqual(len(self.mpm.models), 1)
        self.assertEqual(self.mpm.models[0], self.model)

    def test_n_models(self):
        self.assertEqual(self.mpm.n_models, 1)

    def test_str(self):
        # Add some parameters to the model parameter mapper.
        self.test_def_param()
        try:
            str(self.mpm)
        except:
            self.fail('The __str__ method raised an exception!')

    def test_def_param(self):
        self.mpm.def_param(self.fixed_param)
        self.mpm.def_param(self.floating_param, 'fp')
        self.assertEqual(self.mpm.n_global_params, 2)
        self.assertEqual(self.mpm.n_global_fixed_params, 1)
        self.assertEqual(self.mpm.n_global_floating_params, 1)

    def test_get_model_param_dict(self):
        # Add some parameters to the model parameter mapper.
        self.test_def_param()
        model_param_dict = self.mpm.get_model_param_dict(np.array([2.4]))
        self.assertEqual(len(model_param_dict), 2)
        self.assertTrue('p1' in model_param_dict)
        self.assertTrue('fp' in model_param_dict)
        self.assertAlmostEqual(model_param_dict['p1'], 42)
        self.assertAlmostEqual(model_param_dict['fp'], 2.4)


class MultiModelParameterMapperTestCase(unittest.TestCase):
    def setUp(self):
        self.fixed_param0 = Parameter('p0', 42)
        self.floating_param0 = Parameter('p1', 4, 1, 6)
        self.floating_param1 = Parameter('p2', 13, 10, 15)
        self.model0 = Model('m0')
        self.model1 = Model('m1')
        self.mpm = MultiModelParameterMapper(
            'signal', (self.model0, self.model1))

    def test_name(self):
        self.assertEqual(self.mpm.name, 'signal')

    def test_models(self):
        self.assertEqual(len(self.mpm.models), 2)
        self.assertEqual(self.mpm.models[0], self.model0)
        self.assertEqual(self.mpm.models[1], self.model1)

    def test_n_models(self):
        self.assertEqual(self.mpm.n_models, 2)

    def test_str(self):
        # Add some parameters.
        self.test_def_param()
        try:
            str(self.mpm)
        except:
            self.fail('The __str__ method raised an exception!')

    def test_def_param(self):
        self.mpm.def_param(self.fixed_param0, models=(self.model1,))
        self.mpm.def_param(self.floating_param0, 'fp', models=(self.model0,self.model1))
        self.mpm.def_param(self.floating_param1, models=(self.model1))
        self.assertEqual(self.mpm.n_global_params, 3)
        self.assertEqual(self.mpm.n_global_fixed_params, 1)
        self.assertEqual(self.mpm.n_global_floating_params, 2)

        # The models cannot be an empty set.
        with self.assertRaises(ValueError):
            self.mpm.def_param(self.fixed_param0, 'fp', models=())
        # A model parameter can only be defined once for a given model.
        with self.assertRaises(KeyError):
            self.mpm.def_param(self.fixed_param0, 'fp', models=(self.model0))

    def test_get_model_param_dict(self):
        # Add some parameters to the model parameter mapper.
        self.test_def_param()

        m0_param_dict = self.mpm.get_model_param_dict(
            np.array([2.4, 11.1]), model_idx=0)
        self.assertEqual(len(m0_param_dict), 1)
        self.assertTrue('fp' in m0_param_dict)
        self.assertAlmostEqual(m0_param_dict['fp'], 2.4)

        m1_param_dict = self.mpm.get_model_param_dict(
            np.array([2.4, 11.1]), model_idx=1)
        self.assertEqual(len(m1_param_dict), 3)
        self.assertTrue('p0' in m1_param_dict)
        self.assertTrue('fp' in m1_param_dict)
        self.assertTrue('p2' in m1_param_dict)
        self.assertAlmostEqual(m1_param_dict['fp'], 2.4)
        self.assertAlmostEqual(m1_param_dict['p2'], 11.1)
        self.assertAlmostEqual(m1_param_dict['p0'], 42)


class HypoParameterDefinitionTestCase(unittest.TestCase):
    def setUp(self):
        self.fixed_param0 = Parameter('fixed_param0', 42)
        self.fixed_param1 = Parameter('fixed_param1', 11)
        self.floating_param0 = Parameter('floating_param0', 4, 1, 6)
        self.floating_param1 = Parameter('floating_param1', 0.3, 0.1, 1)

        self.model0 = Model('m0')
        self.model1 = Model('m1')

        self.mpm0 = SingleModelParameterMapper(
            'mpm0', self.model0)
        self.mpm0.def_param(self.fixed_param0)
        self.mpm0.def_param(self.floating_param0, 'fp0')

        self.mpm1 = SingleModelParameterMapper(
            'mpm1', self.model1)
        self.mpm1.def_param(self.fixed_param1)
        self.mpm1.def_param(self.floating_param1, 'fp1')

        self.hpdef = HypoParameterDefinition((self.mpm0, self.mpm1))

    def test_model_param_mapper_list(self):
        mpm_list = self.hpdef.model_param_mapper_list
        self.assertEqual(len(mpm_list), 2)
        self.assertEqual(mpm_list[0], self.mpm0)
        self.assertEqual(mpm_list[1], self.mpm1)

    def test__str__(self):
        try:
            str(self.hpdef)
        except:
            self.fail('The __str__ method raised an exception!')

    def test_copy(self):
        # Check to raise on bad input type.
        with self.assertRaises(TypeError):
            hpdef = self.hpdef.copy('not of type dict')

        # Create an exact copy.
        hpdef = self.hpdef.copy()
        self.assertTrue(isinstance(hpdef, HypoParameterDefinition))
        mpm0 = hpdef['mpm0']
        mpm1 = hpdef['mpm1']
        self.assertTrue((mpm0 is not self.mpm0) and (mpm0 is not self.mpm1))
        self.assertTrue((mpm1 is not self.mpm0) and (mpm1 is not self.mpm1))

        self.assertEqual(mpm0.global_paramset.n_params, 2)
        self.assertEqual(mpm0.global_paramset.n_fixed_params, 1)
        self.assertEqual(mpm0.global_paramset.n_floating_params, 1)
        self.assertTrue(mpm0.global_paramset.params[0] == self.fixed_param0)
        self.assertTrue(mpm0.global_paramset.params[1] == self.floating_param0)

        self.assertEqual(mpm1.global_paramset.n_params, 2)
        self.assertEqual(mpm1.global_paramset.n_fixed_params, 1)
        self.assertEqual(mpm1.global_paramset.n_floating_params, 1)
        self.assertTrue(mpm1.global_paramset.params[0] == self.fixed_param1)
        self.assertTrue(mpm1.global_paramset.params[1] == self.floating_param1)

        self.setUp()

        # Create a copy where we fix one of the floating parameters.
        hpdef = self.hpdef.copy({'floating_param1':2.3})
        self.assertTrue(isinstance(hpdef, HypoParameterDefinition))
        mpm0 = hpdef['mpm0']
        mpm1 = hpdef['mpm1']
        self.assertTrue((mpm0 is not self.mpm0) and (mpm0 is not self.mpm1))
        self.assertTrue((mpm1 is not self.mpm0) and (mpm1 is not self.mpm1))

        self.assertEqual(mpm0.global_paramset.n_params, 2)
        self.assertEqual(mpm0.global_paramset.n_fixed_params, 1)
        self.assertEqual(mpm0.global_paramset.n_floating_params, 1)
        self.assertTrue(mpm0.global_paramset.params[0] == self.fixed_param0)
        self.assertTrue(mpm0.global_paramset.params[1] == self.floating_param0)

        self.assertEqual(mpm1.global_paramset.n_params, 2)
        self.assertEqual(mpm1.global_paramset.n_fixed_params, 2)
        self.assertEqual(mpm1.global_paramset.n_floating_params, 0)
        self.assertTrue(mpm1.global_paramset.params[0] == self.fixed_param1)
        self.assertFalse(mpm1.global_paramset.params[1] == self.floating_param1)
        self.assertTrue(mpm1.global_paramset.params[1].isfixed)
        self.assertAlmostEqual(mpm1.global_paramset.params[1].initial, 2.3)
        self.assertAlmostEqual(mpm1.global_paramset.params[1].value, 2.3)

    def test_create_ParameterSetArray(self):
        paramsetarr = self.hpdef.create_ParameterSetArray()
        self.assertTrue(isinstance(paramsetarr, ParameterSetArray))


class TestParameters(unittest.TestCase):
    def test_MultiSourceFitParameterMapper(self):
        from skyllh.physics.source import PointLikeSource
        from skyllh.core.parameters import (
            FitParameter,
            MultiSourceFitParameterMapper
        )

        # Define a list of point-like sources.
        sources = [
            PointLikeSource(np.deg2rad(120), np.deg2rad(-23)),
            PointLikeSource(np.deg2rad(266), np.deg2rad(61)),
        ]

        # Define the fit parameters 'gamma1' and 'gamma2' which map to the
        # 'gamma' source parameter of the first and second source, respectively.
        sfpm = MultiSourceFitParameterMapper(sources)
        sfpm.def_fit_parameter(FitParameter('gamma1', 1, 4, 2.0), 'gamma', sources[0])
        sfpm.def_fit_parameter(FitParameter('gamma2', 1, 4, 2.1), 'gamma', sources[1])

        # Check the initial values.
        self.assertTrue(np.all(sfpm.fitparamset.initials == np.array([2.0, 2.1])))

        # Get the source parameters for the first source (feed it with the
        # initials).
        fitparams = sfpm.get_src_fitparams(sfpm.fitparamset.initials, 0)
        self.assertEqual(fitparams, {'gamma': 2.0})

        # Get the source parameters for the second source (feed it with the
        # initials).
        fitparams = sfpm.get_src_fitparams(sfpm.fitparamset.initials, 1)
        self.assertEqual(fitparams, {'gamma': 2.1})

if(__name__ == '__main__'):
    unittest.main()



# -*- coding: utf-8 -*-

import os.path

from skyllh.core import session
from skyllh.core.config import CFG

# Automatically enable interactive mode, if the Python interpreter is in
# interactive mode.
if(session.is_python_interpreter_in_interactive_mode()):
    session.enable_interactive_session()
else:
    session.disable_interactive_session()


def wd_filename(filename):
    """Generates the fully qualified file name under the project's working
    directory of the given file.

    Parameters
    ----------
    filename : str
        The name of the file for which to generate the working directory path
        file name.

    Returns
    -------
    pathfilename : str
        The generated fully qualified path file name of ``filename`` with the
        project's working directory prefixed.
    """
    pathfilename = os.path.join(CFG['project']['working_directory'], filename)
    return pathfilename



# -*- coding: utf-8 -*-

"""The ``source`` module contains the base class ``SourceModel`` for modelling a
source in the sky. What kind of properties this source has is modeled by a
derived class. The most common one is the PointLikeSource source model for a
point-like source at a given location in the sky with a given flux model.
"""

import numpy as np

from skyllh.core.py import (
    ObjectCollection,
    classname,
    float_cast,
    issequence
)


class SourceLocation(object):
    """Stores the location of a source, i.e. right-ascention and declination.
    """
    def __init__(self, ra, dec):
        self.ra = ra
        self.dec = dec

    @property
    def ra(self):
        """The right-ascention angle in radian of the source position.
        """
        return self._ra
    @ra.setter
    def ra(self, v):
        v = float_cast(v, 'The ra property must be castable to type float!')
        self._ra = v

    @property
    def dec(self):
        """The declination angle in radian of the source position.
        """
        return self._dec
    @dec.setter
    def dec(self, v):
        v = float_cast(v, 'The dec property must be castable to type float!')
        self._dec = v

class SourceWeights(object):
    """Stores the relative weights of a source, i.e. weights and gradients.
       There are two weights that should be included. one is the detector weight,
       which is declination dependent, and the other is a hypothesis weight, and that
       is provided by the user.
    """
    def __init__(self, src_w=None, src_w_grad=None, src_w_W=None):
        self.src_w      = src_w
        self.src_w_grad = src_w_grad
        self.src_w_W    = src_w_W
    @property
    def src_w(self):
        """The relative weight of the source(s).
        """
        return self._src_w
    @src_w.setter
    def src_w(self, v):
        v = float_cast(v, 'The src_w property must be castable to type float!')
        self._src_w = v

    @property
    def src_w_grad(self):
        """The relative weight gradients of the source(s).
        """
        return self._src_w_grad
    @src_w_grad.setter
    def src_w_grad(self, v):
        v = float_cast(v, 'The src_w_grad property must be castable to type float!')
        self._src_w_grad = v

    @property
    def src_w_W(self):
        """The hypothesis weight of the source(s).
        """
        return self._src_w_W
    @src_w_W.setter
    def src_w_W(self, v):
        v = float_cast(v, 'The src_w_W property must be castable to type float!')
        self._src_w_W = v


class SourceModel(object):
    """The base class for all source models in Skyllh. Each source has a central
    location given by a right-ascention and declination location.
    """
    def __init__(self, ra, dec, src_w=None, src_w_grad=None, src_w_W=None):
        self.loc = SourceLocation(ra, dec)
        src_w = np.ones_like(self.loc.ra, dtype=np.float64)
        src_w_grad = np.zeros_like(self.loc.ra, dtype=np.float64)

        if (src_w_W is None):
            src_w_W = np.ones_like(self.loc.ra, dtype=np.float64)

        self.weight = SourceWeights(src_w, src_w_grad, src_w_W)

    @property
    def loc(self):
        """The location of the source.
        """
        return self._loc
    @loc.setter
    def loc(self, srcloc):
        if(not isinstance(srcloc, SourceLocation)):
            raise TypeError('The loc property must be an instance of SourceLocation!')
        self._loc = srcloc

    @property
    def weight(self):
        """The weight of the source.
        """
        return self._weight
    @weight.setter
    def weight(self, w_src):
        if(not isinstance(w_src, SourceWeights)):
            raise TypeError('The weight property must be an instance of SourceWeights!')
        self._weight = w_src

    @property
    def id(self):
        """(read-only) The ID of the source. It's an integer generated with the
        id() function. Hence, it's related to the memory address of the object.
        """
        return id(self)


class SourceCollection(ObjectCollection):
    """This class describes a collection of sources. It can be used to group
    sources into a single object, for instance for a stacking analysis.
    """
    @staticmethod
    def cast(obj, errmsg):
        """Casts the given object to a SourceCollection object. If the cast
        fails, a TypeError with the given error message is raised.

        Parameters
        ----------
        obj : SourceModel | sequence of SourceModel | SourceCollection
            The object that should be casted to SourceCollection.
        errmsg : str
            The error message if the cast fails.

        Raises
        ------
        TypeError
            If the cast fails.
        """
        if(isinstance(obj, SourceModel)):
            obj = SourceCollection(SourceModel, [obj])
        if(not isinstance(obj, SourceCollection)):
            if(issequence(obj)):
                obj = SourceCollection(SourceModel, obj)
            else:
                raise TypeError(errmsg)
        return obj

    def __init__(self, source_type=None, sources=None):
        """Creates a new source collection.

        Parameters
        ----------
        source_type : type | None
            The type of the source. If set to None (default), SourceModel will
            be used.
        sources : sequence of source_type instances | None
            The sequence of sources this collection should be initalized with.
        """
        if(source_type is None):
            source_type = SourceModel
        super(SourceCollection, self).__init__(sources, obj_type=source_type)

    @property
    def source_type(self):
        """(read-only) The type of the source model.
        """
        return self.obj_type

    @property
    def sources(self):
        """(read-only) The list of sources of type ``source_type``.
        """
        return self.objects


class Catalog(SourceCollection):
    """This class describes a catalog of sources. It is derived from
    SourceCollection. A catalog has a name.
    """
    def __init__(self, name, source_type=None, sources=None):
        """Creates a new source catalog.

        Parameters
        ----------
        name : str
            The name of the catalog.
        source_type : type | None
            The type of the source. If set to None (default), the default type
            defined by SourceCollection will be used.
        sources : sequence of source_type | None
            The sequence of sources this catalog should be initalized with.
        """
        super(Catalog, self).__init__(source_type=source_type, sources=sources)
        self.name = name

    @property
    def name(self):
        """The name of the catalog.
        """
        return self._name
    @name.setter
    def name(self, name):
        if(not isinstance(name, str)):
            raise TypeError('The name property must be of type str!')
        self._name = name

    def as_source_collection(self):
        """Creates a SourceCollection object for this catalog and returns it.
        """
        source_collection = SourceCollection(source_type=self.source_type, sources=self.sources)
        return source_collection


class PointLikeSource(SourceModel):
    """The PointLikeSource class is a source model for a point-like source
    object in the sky at a given location (right-ascention and declination).
    """
    def __init__(self, ra, dec, src_w=None, src_w_grad=None, src_w_W=None):
        super(PointLikeSource, self).__init__(ra, dec, src_w, src_w_grad, src_w_W)

    @property
    def ra(self):
        """(read-only) The right-ascention angle in radian of the source
        position.
        This is a short-cut for `self.loc.ra`.
        """
        return self._loc._ra

    @property
    def dec(self):
        """(read-only) The declination angle in radian of the source position.
        This is a short-cut for `self.loc.dec`.
        """
        return self._loc._dec

    def __str__(self):
        """Pretty string representation of this class instance.
        """
        s = classname(self) + ': { ra=%.3f deg, dec=%.3f deg }'%(
            np.rad2deg(self.ra), np.rad2deg(self.dec))
        return s


class PointLikeSourceCollection(SourceCollection):
    """Describes a collection of point-like sources.
    """
    def __init__(self, sources=None):
        """Creates a new collection of PointLikeSource objects.

        Parameters
        ----------
        sources : sequence of PointLikeSource instances | None
            The sequence of PointLikeSource objects this collection should be
            initalized with.
        """
        super(PointLikeSourceCollection, self).__init__(
            source_type=PointLikeSource, sources=sources)

    @property
    def ra(self):
        """(read-only) The ndarray with the right-ascention of all the sources.
        """
        return np.array([ src.ra for src in self ])

    @property
    def dec(self):
        """(read-only) The ndarray with the declination of all the sources.
        """
        return np.array([ src.dec for src in self ])


class PointLikeSourceCatalog(Catalog):
    """Describes a catalog of point-like sources. The difference to a
    PointLikeSourceCollection is the additional properties of a catalog, e.g.
    the name.
    """
    def __init__(self, name, sources=None):
        """Creates a new point source catalog of the given name.

        Parameters
        ----------
        name : str
            The name of the point-like source catalog.
        sources : sequence of PointLikeSource instances | None
            The sequence of PointLikeSource instances this catalog should be
            initalized with.
        """
        super(PointLikeSourceCatalog, self).__init__(
            name=name, source_type=PointLikeSource, sources=sources)



# -*- coding: utf-8 -*-

from skyllh.core.scrambling import TimeScramblingMethod
from skyllh.i3.coords import hor_to_equ_transform, azi_to_ra_transform


class I3TimeScramblingMethod(TimeScramblingMethod):
    """The I3TimeScramblingMethod class provides a data scrambling method to
    perform time scrambling of the data,
    by drawing a MJD time from a given time generator.
    """
    def __init__(self, timegen):
        """Initializes a new I3 time scrambling instance.

        Parameters
        ----------
        timegen : TimeGenerator
            The time generator that should be used to generate random MJD times.
        """
        super(I3TimeScramblingMethod, self).__init__(timegen, hor_to_equ_transform)

    # We override the scramble method because for IceCube we only need to change
    # the ``ra`` field.
    def scramble(self, rss, data):
        """Draws a time from the time generator and calculates the right
        ascention coordinate from the azimuth angle according to the time.
        Sets the values of the ``time`` and ``ra`` keys of data.

        Parameters
        ----------
        rss : RandomStateService
            The random state service providing the random number
            generator (RNG).
        data : DataFieldRecordArray instance
            The DataFieldRecordArray instance containing the to be scrambled
            data.

        Returns
        -------
        data : numpy record ndarray
            The given numpy record ndarray holding the scrambled data.
        """
        mjds = self._timegen.generate_times(rss, len(data))

        data['time'] = mjds
        data['ra'] = azi_to_ra_transform(data['azi'], mjds)

        return data



# -*- coding: utf-8 -*-

"""This module provides functionality for interpolation.
"""

import abc
import numpy as np

from skyllh.core.parameters import (
    ParameterGrid,
    ParameterGridSet
)
from skyllh.core.py import classname


class GridManifoldInterpolationMethod(object, metaclass=abc.ABCMeta):
    """This is an abstract base class for implementing a method to interpolate
    a manifold that is defined on a grid of parameter values. In general the
    number of parameters can be arbitrary and hence the manifold's
    dimensionality can be arbitrary, too. However, in practice the interpolation
    on a multi-dimensional manifold can be rather difficult.
    Nevertheless, we provide this interface to allow for manifold grids with
    different dimensionality.
    """

    def __init__(self, f, param_grid_set):
        """Constructor for a GridManifoldInterpolationMethod object.
        It must be called by the derived class.

        Parameters
        ----------
        f : callable R^d -> R
            The function that takes d parameters as input and returns the
            value of the d-dimensional manifold at this point for each given
            event.
            The call signature of f must be:

                ``__call__(tdm, gridparams, eventdata)``

            where ``tdm`` is the TrialDataManager instance holding the trial
            data, ``gridparams`` is the dictionary with the parameter values
            on the grid, and ``eventdata`` is a 2-dimensional (N,V)-shaped numpy
            ndarray holding the event data, where N is the number of events, and
            V the dimensionality of the event data.
        param_grid_set : ParameterGrid instance | ParameterGridSet instance
            The set of d parameter grids. This defines the grid of the
            manifold.
        """
        super(GridManifoldInterpolationMethod, self).__init__()

        self.f = f
        self.param_grid_set = param_grid_set

    @property
    def f(self):
        """The R^d -> R manifold function.
        """
        return self._f
    @f.setter
    def f(self, func):
        if(not callable(func)):
            raise TypeError('The f property must be a callable object!')
        self._f = func

    @property
    def param_grid_set(self):
        """The ParameterGridSet object defining the set of d parameter grids.
        This defines the grid of the manifold.
        """
        return self._param_grid_set
    @param_grid_set.setter
    def param_grid_set(self, obj):
        if(isinstance(obj, ParameterGrid)):
            obj = ParameterGridSet([obj])
        elif(not isinstance(obj, ParameterGridSet)):
            raise TypeError('The param_grid_set property must be an instance '
                'of ParameterGridSet!')
        self._param_grid_set = obj

    @property
    def ndim(self):
        """(read-only) The dimensionality of the manifold.
        """
        return len(self._param_grid_set)

    @abc.abstractmethod
    def get_value_and_gradients(self, tdm, eventdata, params):
        """Retrieves the interpolated value of the manifold at the d-dimensional
        point ``params`` for all given events, along with the d gradients,
        i.e. partial derivatives.

        Parameters
        ----------
        tdm : TrialDataManager
            The TrialDataManager instance holding the trial data.
        eventdata : numpy (N_events,V)-shaped 2D ndarray
            The 2D (N_events,V)-shaped numpy ndarray holding the event data,
            where N_events is the number of events, and V the dimensionality of
            the event data.
        params : dict
            The dictionary with the parameter values, defining the point on the
            manifold for which the value should get calculated.

        Returns
        -------
        value : (N,) ndarray of float
            The interpolated manifold value for the N given events.
        gradients : (D,N) ndarray of float
            The D manifold gradients for the N given events, where D is the
            number of parameters. The order of the D parameters is defined
            by the ParameterGridSet that has been provided at construction time
            of this interpolation method object.
        """
        pass


class NullGridManifoldInterpolationMethod(GridManifoldInterpolationMethod):
    """This grid manifold interpolation method performes no interpolation. When
    the ``get_value_and_gradients`` method is called, it rounds the parameter
    values to their nearest grid point values. All gradients are set to zero.
    """
    def __init__(self, f, param_grid_set):
        """Creates a new NullGridManifoldInterpolationMethod instance.

        Parameters
        ----------
        f : callable R^D -> R
            The function that takes the parameter grid value as input and
            returns the value of the n-dimensional manifold at this point for
            each given event.

                ``__call__(tdm, gridparams, eventdata)``

            where ``gridparams`` is the dictionary with the parameter names and
            values on the grid, and ``eventdata`` is a 2-dimensional
            (N,V)-shaped numpy ndarray holding the event data, where N is the
            number of events, and V the dimensionality of the event data.
            The return value of ``f`` must be a (N,)-shaped 1d ndarray of float.
        param_grid_set : ParameterGrid instance | ParameterGridSet instance
            The set of parameter grids. This defines the grid of the
            D-dimensional manifold.
        """
        super(NullGridManifoldInterpolationMethod, self).__init__(
            f, param_grid_set)

    def get_value_and_gradients(self, tdm, eventdata, params):
        """Calculates the non-interpolated manifold value and its gradient
        (zero) for each given event at the point ``params``.
        By definition the D values of ``params`` must coincide with the
        parameter grid values.

        Parameters
        ----------
        tdm : TrialDataManager
            The TrialDataManager instance holding the trial data.
        eventdata : numpy (N_events,V)-shaped 2D ndarray
            The 2D (N_events,V)-shaped numpy ndarray holding the event data,
            where N_events is the number of events, and V the dimensionality of
            the event data.
        params : dict
            The dictionary with the D parameter values, defining the point on
            the manifold for which the value should get calculated.

        Returns
        -------
        value : (N,) ndarray of float
            The interpolated manifold value for the N given events.
        gradients : (D,N) ndarray of float
            The D manifold gradients for the N given events, where D is the
            number of parameters.
            By definition, all gradients are zero.
        """
        # Round the given parameter values to their nearest grid values.
        gridparams = dict()
        for (pname,pvalue) in params.items():
            p_grid = self._param_grid_set[pname]
            gridparams[pname] = p_grid.round_to_nearest_grid_point(pvalue)

        value = self._f(tdm, gridparams, eventdata)
        gradients = np.zeros(
            (len(params), tdm.n_selected_events), dtype=np.float64)

        return (value, gradients)


class Linear1DGridManifoldInterpolationMethod(GridManifoldInterpolationMethod):
    """This grid manifold interpolation method interpolates the 1-dimensional
    grid manifold using a line.
    """
    def __init__(self, f, param_grid_set):
        """Creates a new Linear1DGridManifoldInterpolationMethod instance.

        Parameters
        ----------
        f : callable R -> R
            The function that takes the parameter grid value as input and
            returns the value of the 1-dimensional manifold at this point for
            each given event.

                ``__call__(tdm, gridparams, eventdata)``

            where ``gridparams`` is the dictionary with the parameter names and
            values on the grid, and ``eventdata`` is a 2-dimensional
            (N,V)-shaped numpy ndarray holding the event data, where N is the
            number of events, and V the dimensionality of the event data.
            The return value of ``f`` must be a (N,)-shaped 1d ndarray of float.
        param_grid_set : ParameterGrid instance | ParameterGridSet instance
            The set of parameter grids. This defines the grid of the
            1-dimensional manifold. By definition, only the first parameter grid
            is considered.
        """
        super(Linear1DGridManifoldInterpolationMethod, self).__init__(
            f, param_grid_set)

        if(len(self._param_grid_set) != 1):
            raise ValueError('The %s class supports only 1D grid manifolds. '
                'The param_grid_set argument must contain 1 ParameterGrid '
                'instance! Currently it has %d!'%(
                    classname(self), len(self._param_grid_set)))
        self.p_grid = self._param_grid_set[0]

        # Create a cache for the line parameterization for the last
        # manifold grid point for the different events.
        self._create_cache(None, np.array([]), np.array([]))
        self._cache_tdm_trial_data_state_id = None

    def _create_cache(self, x0, m, b):
        """Creates a cache for the line parameterization for the last manifold
        grid point for the nevents different events.

        Parameters
        ----------
        x0 : float | None
            The parameter grid value for the lower point of the grid manifold
            used to estimate the line.
        m : 1d ndarray
            The slope of the line for each event.
        b : 1d ndarray
            The offset coefficient of the line for each event.
        """
        self._cache = {
            'x0': x0,
            'm': m,
            'b': b
        }

    def get_value_and_gradients(self, tdm, eventdata, params):
        """Calculates the interpolated manifold value and its gradient for each
        given event at the point ``params``.

        Parameters
        ----------
        tdm : TrialDataManager
            The TrialDataManager instance holding the trial data.
        eventdata : numpy (N_events,V)-shaped 2D ndarray
            The 2D (N_events,V)-shaped numpy ndarray holding the event data,
            where N_events is the number of events, and V the dimensionality of
            the event data.
        params : dict
            The dictionary with the parameter values, defining the point on the
            manifold for which the value should get calculated.

        Returns
        -------
        value : (N,) ndarray of float
            The interpolated manifold value for the N given events.
        gradients : (D,N) ndarray of float
            The D manifold gradients for the N given events, where D is the
            number of parameters.
        """
        (xname, x) = tuple(params.items())[0]

        self__p_grid = self.p_grid

        # Determine the nearest grid point that is lower than x and use that as
        # x0.
        x0 = self__p_grid.round_to_lower_grid_point(x)

        # Check if the line parametrization for x0 is already cached.
        self__cache = self._cache

        tdm_trial_data_state_id = tdm.trial_data_state_id
        cache_tdm_trial_data_state_id = self._cache_tdm_trial_data_state_id

        if((self__cache['x0'] == x0) and
           (tdm.n_selected_events == len(self__cache['m'])) and
           (cache_tdm_trial_data_state_id is not None) and
           (cache_tdm_trial_data_state_id == tdm_trial_data_state_id)
          ):
            m = self__cache['m']
            b = self__cache['b']
        else:
            # Calculate the line parametrization for all the given events.
            self__f = self.f

            # Calculate the upper grid point of x.
            x1 = self__p_grid.round_to_upper_grid_point(x)

            # Check if x was on a grid point. In that case x0 and x1 are equal.
            # The value will be of that grid point x0, but the gradient is
            # calculated based on the two neighboring grid points of x0.
            if(x1 == x0):
                value = self__f(tdm, {xname:x0}, eventdata)
                x0 = self__p_grid.round_to_nearest_grid_point(
                    x0 - self__p_grid.delta)
                x1 = self__p_grid.round_to_nearest_grid_point(
                    x1 + self__p_grid.delta)

                M0 = self__f(tdm, {xname:x0}, eventdata)
                M1 = self__f(tdm, {xname:x1}, eventdata)
                m = (M1 - M0) / (x1 - x0)
                return (value, np.atleast_2d(m))

            M0 = self__f(tdm, {xname:x0}, eventdata)
            M1 = self__f(tdm, {xname:x1}, eventdata)

            m = (M1 - M0) / (x1 - x0)
            b = M0 - m*x0

            # Cache the line parametrization.
            self._create_cache(x0, m, b)
            self._cache_tdm_trial_data_state_id = tdm_trial_data_state_id

        # Calculate the interpolated manifold value. The gradient is m.
        value = m*x + b

        return (value, np.atleast_2d(m))


class Parabola1DGridManifoldInterpolationMethod(GridManifoldInterpolationMethod):
    """This grid manifold interpolation method interpolates the 1-dimensional
    grid manifold using a parabola.
    """
    def __init__(self, f, param_grid_set):
        """Creates a new Parabola1DGridManifoldInterpolationMethod instance.

        Parameters
        ----------
        f : callable R -> R
            The function that takes the parameter grid value as input and
            returns the value of the 1-dimensional manifold at this point for
            each given event.
            The call signature of f must be:

                ``__call__(tdm, gridparams, eventdata)``

            where ``gridparams`` is the dictionary with the parameter names and
            values on the grid, and ``eventdata`` is a 2-dimensional
            (N,V)-shaped numpy ndarray holding the event data, where N is the
            number of events, and V the dimensionality of the event data.
        param_grid_set : instance of ParameterGridSet
            The set of parameter grids. This defines the grid of the
            1-dimensional manifold. By definition, only the first parameter grid
            is considered.
        """
        super(Parabola1DGridManifoldInterpolationMethod, self).__init__(
            f, param_grid_set)

        if(len(self._param_grid_set) != 1):
            raise ValueError('The %s class supports only 1D grid manifolds. '
                'The param_grid_set argument must contain 1 ParameterGrid '
                'instance! Currently it has %d!'%(
                    classname(self), len(self._param_grid_set)))
        self._p_grid = self._param_grid_set[0]

        # Create a cache for the parabola parameterization for the last
        # manifold grid point for the different events.
        self._create_cache(None, np.array([]), np.array([]), np.array([]))
        self._cache_tdm_trial_data_state_id = None

    def _create_cache(self, x1, M1, a, b):
        """Creates a cache for the parabola parameterization for the last
        manifold grid point for the nevents different events.

        Parameters
        ----------
        x1 : float | None
            The parameter grid value for middle point of the grid manifold used
            to estimate the parabola.
        M1 : 1d ndarray
            The grid manifold value for each event of the middle point (x1,).
        a : 1d ndarray
            The parabola coefficient ``a`` for each event.
        b : 1d ndarray
            The parabola coefficient ``b`` for each event.
        """
        self._cache = {
            'x1': x1,
            'M1': M1,
            'a': a,
            'b': b
        }

    def get_value_and_gradients(self, tdm, eventdata, params):
        """Calculates the interpolated manifold value and its gradient for each
        given event at the point ``params``.

        Parameters
        ----------
        tdm : TrialDataManager
            The TrialDataManager instance holding the trial data.
        eventdata : numpy (N_events,V)-shaped 2D ndarray
            The 2D (N_events,V)-shaped numpy ndarray holding the event data,
            where N_events is the number of events, and V the dimensionality of
            the event data.
        params : dict
            The dictionary with the parameter values, defining the point on the
            manifold for which the value should get calculated.

        Returns
        -------
        value : (N,) ndarray of float
            The interpolated manifold value for the N given events.
        gradients : (D,N) ndarray of float
            The D manifold gradients for the N given events, where D is the
            number of parameters.
        """
        (xname, x) = tuple(params.items())[0]

        # Create local variable name alias to avoid Python dot lookups.
        self__p_grid = self._p_grid
        self__p_grid__round_to_nearest_grid_point = \
            self__p_grid.round_to_nearest_grid_point
        self__cache = self._cache

        tdm_trial_data_state_id = tdm.trial_data_state_id
        cache_tdm_trial_data_state_id = self._cache_tdm_trial_data_state_id

        # Determine the nearest grid point x1.
        x1 = self__p_grid__round_to_nearest_grid_point(x)

        # Check if the parabola parametrization for x1 is already cached.
        if((self__cache['x1'] == x1) and
           (tdm.n_selected_events == len(self__cache['M1'])) and
           (cache_tdm_trial_data_state_id is not None) and
           (cache_tdm_trial_data_state_id == tdm_trial_data_state_id)
          ):
            M1 = self__cache['M1']
            a = self__cache['a']
            b = self__cache['b']
        else:
            dx = self__p_grid.delta

            # Calculate the neighboring grid points to x1: x0 and x2.
            x0 = self__p_grid__round_to_nearest_grid_point(x1 - dx)
            x2 = self__p_grid__round_to_nearest_grid_point(x1 + dx)

            # Parameterize the parabola with parameters a, b, and M1.
            self__f = self.f
            M0 = self__f(tdm, {xname:x0}, eventdata)
            M1 = self__f(tdm, {xname:x1}, eventdata)
            M2 = self__f(tdm, {xname:x2}, eventdata)

            a = 0.5*(M0 - 2.*M1 + M2) / dx**2
            b = 0.5*(M2 - M0) / dx

            # Cache the parabola parametrization.
            self._create_cache(x1, M1, a, b)
            self._cache_tdm_trial_data_state_id = tdm_trial_data_state_id

        # Calculate the interpolated manifold value.
        value = a * (x - x1)**2 + b * (x - x1) + M1
        # Calculate the gradient of the manifold.
        gradients = 2. * a * (x - x1) + b

        return (value, np.atleast_2d(gradients))




# -*- coding: utf-8 -*-
import numpy as np

from skyllh.core.py import issequenceof
from skyllh.core.detsigyield import DetSigYieldImplMethod
from skyllh.core.signal_generation import SignalGenerationMethod
from skyllh.physics.source import SourceModel
from skyllh.physics.flux import FluxModel


class SourceHypoGroup(object):
    """The source hypothesis group class provides a data container to describe
    a group of sources that share the same flux model, detector signal yield,
    and signal generation implementation methods.
    """
    def __init__(
            self, sources, fluxmodel, detsigyield_implmethods,
            sig_gen_method=None, source_weights=None):
        """Constructs a new source hypothesis group.

        Parameters
        ----------
        sources : SourceModel | sequence of SourceModel
            The source or sequence of sources that define the source group.
        fluxmodel : instance of FluxModel
            The FluxModel instance that applies to the list of sources of the
            group.
        detsigyield_implmethods : sequence of DetSigYieldImplMethod instances
            The sequence of detector signal yield implementation method
            instances, which should be used to create the detector signal
            yield for the sources of this group. Each element is the
            detector signal yield implementation method for the particular
            dataset, if several datasets are used. If this list contains only
            one implementation method, it should be used for all datasets.
        sig_gen_method : SignalGenerationMethod instance | None
            The instance of SignalGenerationMethod that implements the signal
            generation for the specific detector and source hypothesis. It can
            be set to None, which means, no signal can be generated. Useful for
            data unblinding and data trial generation, where no signal is
            required.
        source_weights : float | sequence of floats | None
            The sequence of relative source weights, normalized to 1.
        """
        self.source_list = sources
        self.fluxmodel = fluxmodel
        self.detsigyield_implmethod_list = detsigyield_implmethods
        self.sig_gen_method = sig_gen_method
        self.source_weights = source_weights

    @property
    def source_list(self):
        """The list of SourceModel instances for which the group is defined.
        """
        return self._source_list
    @source_list.setter
    def source_list(self, sources):
        if(isinstance(sources, SourceModel)):
            sources = [ sources ]
        if(not issequenceof(sources, SourceModel)):
            raise TypeError('The source_list property must be an instance of SourceModel or a sequence of SourceModel instances!')
        self._source_list = list(sources)

    @property
    def fluxmodel(self):
        """The FluxModel instance that applies to the list of sources of this
        source group.
        """
        return self._fluxmodel
    @fluxmodel.setter
    def fluxmodel(self, fluxmodel):
        if(not isinstance(fluxmodel, FluxModel)):
            raise TypeError('The fluxmodel property must be an instance of '
                'FluxModel!')
        self._fluxmodel = fluxmodel

    @property
    def detsigyield_implmethod_list(self):
        """The list of DetSigYieldImplMethod instances, which should be used to
        create the detector signal yield for this group of sources. Each
        element is the detector signal yield implementation method for
        the particular dataset, if several datasets are used. If this list
        contains only one implementation method, it should be used for all
        datasets.
        """
        return self._detsigyield_implmethod_list
    @detsigyield_implmethod_list.setter
    def detsigyield_implmethod_list(self, methods):
        if(isinstance(methods, DetSigYieldImplMethod)):
            methods = [ methods ]
        if(not issequenceof(methods, DetSigYieldImplMethod)):
            raise TypeError('The detsigyield_implmethod_list property must be '
                'a sequence of DetSigYieldImplMethod instances!')
        self._detsigyield_implmethod_list = methods

    @property
    def sig_gen_method(self):
        """The instance of SignalGenerationMethod that implements the signal
        generation for the specific detector and source hypothesis. It can
        be None, which means, no signal can be generated. Useful for
        data unblinding and data trial generation, where no signal is
        required.
        """
        return self._sig_gen_method
    @sig_gen_method.setter
    def sig_gen_method(self, method):
        if(method is not None):
            if(not isinstance(method, SignalGenerationMethod)):
                raise TypeError('The sig_gen_method property must be an '
                    'instance of SignalGenerationMethod!')
        self._sig_gen_method = method

    @property
    def source_weights(self):
        """The 1d array of relative source weights.
        """
        return self._source_weights
    @source_weights.setter
    def source_weights(self, source_weights):
        if(source_weights is None):
            self._source_weights = source_weights
        else:
            if(isinstance(source_weights, (int, float))):
                source_weights = [source_weights]
            if(not issequenceof(source_weights, (int, float))):
                raise TypeError(
                    'The source_weights property must be a sequence of floats!')
            if not(1.0 - 1e-3 <= np.sum(source_weights) <= 1.0 + 1e-3):
                raise ValueError(
                    'The sum of source_weights has to be equal to 1!')
            self._source_weights = np.array(source_weights)

    @property
    def n_sources(self):
        """(read-only) The number of sources within this source hypothesis
        group.
        """
        return len(self._source_list)



# -*- coding: utf-8 -*-

"""The ``signalpdf`` module contains possible signal PDF models for the
likelihood function.
"""

import numpy as np
import scipy as scp

from skyllh.core import display
from skyllh.core.py import (
    classname,
    issequenceof
)
from skyllh.core.livetime import Livetime
from skyllh.core.pdf import (
    PDFAxis,
    IsSignalPDF,
    MultiDimGridPDF,
    MultiDimGridPDFSet,
    MappedMultiDimGridPDFSet,
    NDPhotosplinePDF,
    SpatialPDF,
    TimePDF
)
from skyllh.core.source_hypothesis import SourceHypoGroupManager
from skyllh.physics.source import PointLikeSource
from skyllh.physics.time_profile import TimeProfileModel


class GaussianPSFPointLikeSourceSignalSpatialPDF(SpatialPDF, IsSignalPDF):
    """This spatial signal PDF model describes the spatial PDF for a point
    source smeared with a 2D gaussian point-spread-function (PSF).
    Mathematically, it's the convolution of a point in the sky, i.e. the source
    location, with the PSF. The result of this convolution has the gaussian form

        1/(2*\pi*\sigma^2) * exp(-1/2*(r / \sigma)**2),

    where \sigma is the spatial uncertainty of the event and r the distance on
    the sphere between the source and the data event.

    This PDF requires the `src_array` data field, that is numpy record ndarray
    with the data fields `ra` and `dec` holding the right-ascention and
    declination of the point-like sources, respectively.
    """

    def __init__(self, ra_range=None, dec_range=None, **kwargs):
        """Creates a new spatial signal PDF for point-like sources with a
        gaussian point-spread-function (PSF).

        Parameters
        ----------
        ra_range : 2-element tuple | None
            The range in right-ascention this spatial PDF is valid for.
            If set to None, the range (0, 2pi) is used.
        dec_range : 2-element tuple | None
            The range in declination this spatial PDF is valid for.
            If set to None, the range (-pi/2, +pi/2) is used.
        """
        if(ra_range is None):
            ra_range = (0, 2*np.pi)
        if(dec_range is None):
            dec_range = (-np.pi/2, np.pi/2)

        super(GaussianPSFPointLikeSourceSignalSpatialPDF, self).__init__(
            ra_range=ra_range,
            dec_range=dec_range,
            **kwargs)

    def get_prob(self, tdm, fitparams=None, tl=None):
        """Calculates the spatial signal probability of each event for all given
        sources.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance holding the trial event data for which
            to calculate the PDF values. The following data fields need to be
            present:

            'src_array' : numpy record ndarray
                The numpy record ndarray with the following data fields:

                `ra`: float
                    The right-ascention of the point-like source.
                `dec`: float
                    The declination of the point-like source.

            'ra' : float
                The right-ascention in radian of the data event.
            'dec' : float
                The declination in radian of the data event.
            'ang_err': float
                The reconstruction uncertainty in radian of the data event.
        fitparams : None
            Unused interface argument.
        tl : TimeLord instance | None
            The optional TimeLord instance to use for measuring timing
            information.

        Returns
        -------
        prob : (N_sources,N_events) shaped 2D ndarray
            The ndarray holding the spatial signal probability on the sphere for
            each source and event.
        """
        get_data = tdm.get_data
        src_ev_idxs = tdm.src_ev_idxs

        ra = get_data('ra')
        dec = get_data('dec')
        sigma = get_data('ang_err')

        if len(ra) == 1:
            self.param_set = None

        try:
            # angular difference is pre calculated
            prob = get_data('spatial_pdf_gauss')
            src_ra = get_data('src_array')['ra']

            if src_ev_idxs is None:
                prob = prob.reshape((len(get_data('src_array')), len(ra)))
            else:
                (src_idxs, ev_idxs) = src_ev_idxs
                sigma = np.take(sigma, src_ev_idxs[1])

        except:
            # psi is calculated here
            if src_ev_idxs is None:
                # Make the source position angles two-dimensional so the PDF value can
                # be calculated via numpy broadcasting automatically for several
                # sources. This is useful for stacking analyses.
                src_ra = get_data('src_array')['ra'][:, np.newaxis]
                src_dec = get_data('src_array')['dec'][:, np.newaxis]

                delta_dec = np.abs(dec - src_dec)
                delta_ra = np.abs(ra - src_ra)
                x = (np.sin(delta_dec / 2.))**2. + np.cos(dec) *\
                    np.cos(src_dec) * (np.sin(delta_ra / 2.))**2.
            else:
                # Calculate the angular difference only for events that are close
                # to the respective source poisition. This is useful for stacking
                # analyses.
                (src_idxs, ev_idxs) = src_ev_idxs
                src_ra = get_data('src_array')['ra'][src_idxs]
                src_dec = get_data('src_array')['dec'][src_idxs]

                delta_dec = np.abs(np.take(dec, ev_idxs) - src_dec)
                delta_ra = np.abs(np.take(ra, ev_idxs) - src_ra)
                x = (np.sin(delta_dec / 2.))**2. + np.cos(np.take(dec, ev_idxs)) *\
                    np.cos(src_dec) * (np.sin(delta_ra / 2.))**2.

                # also extend the sigma array to account for all relevant events
                sigma = np.take(sigma, ev_idxs)

                # Handle possible floating precision errors.
            x[x < 0.] = 0.
            x[x > 1.] = 1.

            psi = (2.0*np.arcsin(np.sqrt(x)))

            prob = 0.5/(np.pi*sigma**2)*np.exp(-0.5*(psi/sigma)**2)

        # If the signal hypothesis contains single source
        # return the output here.
        if(len(get_data('src_array')['ra']) == 1):
            grads = np.array([], dtype=np.float64)
            # The new interface returns the pdf only for a single source.
            return (prob[0], grads)
        else:
            # If the signal hypothesis contains multiple sources convolve
            # the pdfs with the source weights.
            src_w = get_data('src_array')['src_w'] * tdm.get_data('src_array')['src_w_W']
            src_w_grads = get_data('src_array')['src_w_grad'] * tdm.get_data('src_array')['src_w_W']

            norm = src_w.sum()
            src_w /= norm
            src_w_grads /= norm

            if src_ev_idxs is not None:
                prob = scp.sparse.csr_matrix((prob, (ev_idxs, src_idxs)))
            else:
                prob = prob.T
            prob_res = prob.dot(src_w)
            grads = (prob.dot(src_w_grads) -
                     prob_res*src_w_grads.sum())

            return (prob_res, np.atleast_2d(grads))


class SignalTimePDF(TimePDF, IsSignalPDF):
    """This class provides a time PDF class for a signal source. It consists of
    a Livetime instance and a TimeProfileModel instance. Together they construct
    the actual signal time PDF, which has detector down-time taking into
    account.
    """

    def __init__(self, livetime, time_profile):
        """Creates a new signal time PDF instance for a given time profile of
        the source.

        Parameters
        ----------
        livetime : Livetime instance
            An instance of Livetime, which provides the detector live-time
            information.
        time_profile : TimeProfileModel instance
            The time profile of the source.
        """
        super(SignalTimePDF, self).__init__()

        self.livetime = livetime
        self.time_profile = time_profile

        # Define the time axis with the time boundaries of the live-time.
        self.add_axis(PDFAxis(
            name='time',
            vmin=self._livetime.time_window[0],
            vmax=self._livetime.time_window[1]))

        # Get the total integral, I, of the time profile and the sum, S, of the
        # integrals for each detector on-time interval during the time profile,
        # in order to be able to rescale the time profile to unity with
        # overlapping detector off-times removed.
        (self._I, self._S) = self._calculate_time_profile_I_and_S()

    @property
    def livetime(self):
        """The instance of Livetime, which provides the detector live-time
        information.
        """
        return self._livetime

    @livetime.setter
    def livetime(self, lt):
        if(not isinstance(lt, Livetime)):
            raise TypeError(
                'The livetime property must be an instance of Livetime!')
        self._livetime = lt

    @property
    def time_profile(self):
        """The instance of TimeProfileModel providing the (assumed) physical
        time profile of the source.
        """
        return self._time_profile

    @time_profile.setter
    def time_profile(self, tp):
        if(not isinstance(tp, TimeProfileModel)):
            raise TypeError(
                'The time_profile property must be an instance of '
                'TimeProfileModel!')
        self._time_profile = tp

    def __str__(self):
        """Pretty string representation of the signal time PDF.
        """
        s = '%s(\n' % (classname(self))
        s += ' '*display.INDENTATION_WIDTH + \
            'livetime = %s,\n' % (str(self._livetime))
        s += ' '*display.INDENTATION_WIDTH + \
            'time_profile = %s\n' % (str(self._time_profile))
        s += ')'
        return s

    def _calculate_time_profile_I_and_S(self):
        """Calculates the total integral, I, of the time profile and the sum, A,
        of the time-profile integrals during the detector on-time intervals.

        Returns
        -------
        I : float
            The total integral of the source time-profile.
        S : float
            The sum of the source time-profile integrals during the detector
            on-time intervals.
        """
        ontime_intervals = self._livetime.get_ontime_intervals_between(
            self._time_profile.t_start, self._time_profile.t_end)
        I = self._time_profile.get_total_integral()
        S = np.sum(self._time_profile.get_integral(
            ontime_intervals[:, 0], ontime_intervals[:, 1]))
        return (I, S)

    def assert_is_valid_for_exp_data(self, data_exp):
        """Checks if the time PDF is valid for all the given experimental data.
        It checks if the time of all events is within the defined time axis of
        the PDF.

        Parameters
        ----------
        data_exp : numpy record ndarray
            The array holding the experimental data. The following data fields
            must exist:

            - 'time' : float
                The MJD time of the data event.

        Raises
        ------
        ValueError
            If some of the data is outside the time range of the PDF.
        """
        time_axis = self.get_axis('time')

        if(np.any((data_exp['time'] < time_axis.vmin) |
                  (data_exp['time'] > time_axis.vmax))):
            raise ValueError('Some data is outside the time range (%.3f, %.3f)!' % (
                time_axis.vmin, time_axis.vmax))

    def get_prob(self, tdm, fitparams):
        """Calculates the signal time probability of each event for the given
        set of signal time fit parameter values.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager holding the trial event data for
            which to calculate the PDF value. The following data fields must
            exist:

            - 'time' : float
                The MJD time of the event.
        fitparams : dict
            The dictionary holding the signal time parameter values for which
            the signal time probability should be calculated.

        Returns
        -------
        prob : array of float
            The (N,)-shaped ndarray holding the probability for each event.
        """
        # Update the time-profile if its fit-parameter values have changed and
        # recalculate self._I and self._S if an updated was actually performed.
        updated = self._time_profile.update(fitparams)
        if(updated):
            (self._I, self._S) = self._calculate_time_profile_I_and_S()

        events_time = tdm.get_data('time')

        # Get a mask of the event times which fall inside a detector on-time
        # interval.
        on = self._livetime.is_on(events_time)

        # The sum of the on-time integrals of the time profile, A, will be zero
        # if the time profile is entirly during detector off-time.
        prob = np.zeros((tdm.n_selected_events,), dtype=np.float64)
        if(self._S > 0):
            prob[on] = self._time_profile.get_value(
                events_time[on]) / (self._I * self._S)

        return prob


class SignalMultiDimGridPDF(MultiDimGridPDF, IsSignalPDF):
    """This class provides a multi-dimensional signal PDF. The PDF is created
    from pre-calculated PDF data on a grid. The grid data is interpolated using
    a :class:`scipy.interpolate.RegularGridInterpolator` instance.
    """

    def __init__(self, axis_binnings, path_to_pdf_splinetable=None,
                 pdf_grid_data=None, norm_factor_func=None):
        """Creates a new signal PDF instance for a multi-dimensional PDF given
        as PDF values on a grid. The grid data is interpolated with a
        :class:`scipy.interpolate.RegularGridInterpolator` instance. As grid
        points the bin edges of the axis binning definitions are used.

        Parameters
        ----------
        axis_binnings : sequence of BinningDefinition
            The sequence of BinningDefinition instances defining the binning of
            the PDF axes. The name of each BinningDefinition instance defines
            the event field name that should be used for querying the PDF.
        path_to_pdf_splinetable : str
            The path to the file containing the spline table.
            The spline table contains a  pre-computed fit to pdf_grid_data.
        pdf_grid_data : n-dimensional numpy ndarray
            The n-dimensional numpy ndarray holding the PDF values at given grid
            points. The grid points must match the bin edges of the given
            BinningDefinition instances of the `axis_binnings` argument.
        norm_factor_func : callable | None
            The function that calculates a possible required normalization
            factor for the PDF value based on the event properties.
            The call signature of this function
            must be `__call__(pdf, events, fitparams)`, where `pdf` is this PDF
            instance, `events` is a numpy record ndarray holding the events for
            which to calculate the PDF values, and `fitparams` is a dictionary
            with the current fit parameter names and values.
        """
        super(SignalMultiDimGridPDF, self).__init__(
            axis_binnings=axis_binnings,
            path_to_pdf_splinetable=path_to_pdf_splinetable,
            pdf_grid_data=pdf_grid_data,
            norm_factor_func=norm_factor_func)


class SignalMultiDimGridPDFSet(MultiDimGridPDFSet, IsSignalPDF):
    """This class extends the MultiDimGridPDFSet PDF class to be a signal PDF.
    See the documentation of the :class:`skyllh.core.pdf.MultiDimGridPDFSet`
    class for what this PDF provides.
    """

    def __init__(self, param_set, param_grid_set, gridparams_pdfs,
                 interpolmethod=None, **kwargs):
        """Creates a new SignalMultiDimGridPDFSet instance, which holds a set of
        MultiDimGridPDF instances, one for each point of a parameter grid set.

        Parameters
        ----------
        param_set : Parameter instance | sequence of Parameter instances |
                    ParameterSet instance
            The set of parameters defining the model parameters of this PDF.
        param_grid_set : ParameterGrid instance | ParameterGridSet instance
            The set of ParameterGrid instances, which define the grid values of
            the model parameters, the given MultiDimGridPDF instances belong to.
        gridparams_pdfs : sequence of (dict, MultiDimGridPDF) tuples
            The sequence of 2-element tuples which define the mapping of grid
            values to PDF instances.
        interpolmethod : subclass of GridManifoldInterpolationMethod
            The class specifying the interpolation method. This must be a
            subclass of ``GridManifoldInterpolationMethod``.
            If set to None, the default grid manifold interpolation method
            ``Linear1DGridManifoldInterpolationMethod`` will be used.
        """
        super(SignalMultiDimGridPDFSet, self).__init__(
            param_set=param_set,
            param_grid_set=param_grid_set,
            gridparams_pdfs=gridparams_pdfs,
            interpolmethod=interpolmethod,
            pdf_type=SignalMultiDimGridPDF,
            **kwargs)


class SignalMappedMultiDimGridPDFSet(MappedMultiDimGridPDFSet, IsSignalPDF):
    """This class extends the MappedMultiDimGridPDFSet PDF class to be a signal
    PDF. See the documentation of the
    :class:`skyllh.core.pdf.MappedMultiDimGridPDFSet` class for what this PDF
    provides.
    """

    def __init__(self, param_grid_set, gridparams_pdfs,
                 interpolmethod=None, **kwargs):
        """Creates a new SignalMappedMultiDimGridPDFSet instance, which holds a
        set of MultiDimGridPDF instances, one for each point of a parameter grid
        set.

        Parameters
        ----------
        param_grid_set : ParameterGrid instance | ParameterGridSet instance
            The set of ParameterGrid instances, which define the grid values of
            the model parameters, the given MultiDimGridPDF instances belong to.
        gridparams_pdfs : sequence of (dict, MultiDimGridPDF) tuples
            The sequence of 2-element tuples which define the mapping of grid
            values to PDF instances.
        """
        super(SignalMappedMultiDimGridPDFSet, self).__init__(
            param_grid_set=param_grid_set,
            gridparams_pdfs=gridparams_pdfs,
            pdf_type=SignalMultiDimGridPDF,
            **kwargs)


class SignalNDPhotosplinePDF(NDPhotosplinePDF, IsSignalPDF):
    """This class provides a multi-dimensional signal PDF created from a
    n-dimensional photospline fit. The photospline package is used to evaluate
    the PDF fit.
    """

    def __init__(
            self,
            axis_binnings,
            param_set,
            path_to_pdf_splinefit,
            norm_factor_func=None):
        """Creates a new signal PDF instance for a n-dimensional photospline PDF
        fit.

        Parameters
        ----------
        axis_binnings : BinningDefinition | sequence of BinningDefinition
            The sequence of BinningDefinition instances defining the binning of
            the PDF axes. The name of each BinningDefinition instance defines
            the event field name that should be used for querying the PDF.
        param_set : Parameter | ParameterSet
            The Parameter instance or ParameterSet instance defining the
            parameters of this PDF. The ParameterSet holds the information
            which parameters are fixed and which are floating (i.e. fitted).
        path_to_pdf_splinefit : str
            The path to the file containing the photospline fit.
        norm_factor_func : callable | None
            The function that calculates a possible required normalization
            factor for the PDF value based on the event properties.
            The call signature of this function must be
            `__call__(pdf, tdm, params)`, where `pdf` is this PDF
            instance, `tdm` is an instance of TrialDataManager holding the
            event data for which to calculate the PDF values, and `params` is a
            dictionary with the current parameter names and values.
        """
        super(SignalNDPhotosplinePDF, self).__init__(
            axis_binnings=axis_binnings,
            param_set=param_set,
            path_to_pdf_splinefit=path_to_pdf_splinefit,
            norm_factor_func=norm_factor_func
        )



# -*- coding: utf-8 -*-

"""This module contains classes for IceCube specific detector signal yields,
for a variation of source model and flux model combinations.
"""

import abc
import numpy as np

import scipy.interpolate

from skyllh.core import multiproc
from skyllh.core.py import issequenceof
from skyllh.core.binning import BinningDefinition
from skyllh.core.parameters import ParameterGrid
from skyllh.core.detsigyield import (
    DetSigYield,
    DetSigYieldImplMethod,
    get_integrated_livetime_in_days
)
from skyllh.core.livetime import Livetime
from skyllh.physics.source import PointLikeSource
from skyllh.physics.flux import (
    FluxModel,
    PowerLawFlux,
    get_conversion_factor_to_internal_flux_unit
)


class I3DetSigYield(DetSigYield, metaclass=abc.ABCMeta):
    """Abstract base class for all IceCube specific detector signal yield
    classes. It assumes that sin(dec) binning is required for calculating the
    detector effective area and hence the detector signal yield.
    """

    def __init__(self, implmethod, dataset, fluxmodel, livetime, sin_dec_binning):
        """Constructor of the IceCube specific detector signal yield base
        class.
        """
        super(I3DetSigYield, self).__init__(implmethod, dataset, fluxmodel, livetime)

        self.sin_dec_binning = sin_dec_binning

    @property
    def sin_dec_binning(self):
        """The BinningDefinition instance defining the sin(dec) binning
        definition.
        """
        return self._sin_dec_binning
    @sin_dec_binning.setter
    def sin_dec_binning(self, bd):
        if(not isinstance(bd, BinningDefinition)):
            raise TypeError('The sin_dec_binning property must be an instance '
                'of BinningDefinition!')
        self._sin_dec_binning = bd


class I3DetSigYieldImplMethod(DetSigYieldImplMethod, metaclass=abc.ABCMeta):
    """Abstract base class for an IceCube specific detector signal yield
    implementation method class.
    """

    def __init__(self, sin_dec_binning=None, **kwargs):
        """Constructor of the IceCube specific detector signal yield
        implementation base class.

        Parameters
        ----------
        sin_dec_binning : BinningDefinition instance
            The instance of BinningDefinition defining the binning of sin(dec).
        """
        super(I3DetSigYieldImplMethod, self).__init__(**kwargs)

        self.sin_dec_binning = sin_dec_binning

    @property
    def sin_dec_binning(self):
        """The BinningDefinition instance for the sin(dec) binning that should
        be used for computing the sin(dec) dependency of the detector signal
        yield. If None, the binning is supposed to be taken from the Dataset's
        binning definitions.
        """
        return self._sin_dec_binning
    @sin_dec_binning.setter
    def sin_dec_binning(self, binning):
        if((binning is not None) and
           (not isinstance(binning, BinningDefinition))):
            raise TypeError('The sin_dec_binning property must be None, or '
                'an instance of BinningDefinition!')
        self._sin_dec_binning = binning

    def get_sin_dec_binning(self, dataset):
        """Gets the sin(dec) binning definition either as setting from this
        detector signal yield implementation method itself, or from the
        given dataset.
        """
        sin_dec_binning = self.sin_dec_binning
        if(sin_dec_binning is None):
            if(not dataset.has_binning_definition('sin_dec')):
                raise KeyError('No binning definition named "sin_dec" is '
                    'defined in the dataset and no user defined binning '
                    'definition was provided to this detector signal yield '
                    'implementation method!')
            sin_dec_binning = dataset.get_binning_definition('sin_dec')
        return sin_dec_binning


class PointLikeSourceI3DetSigYieldImplMethod(
        I3DetSigYieldImplMethod, metaclass=abc.ABCMeta):
    """Abstract base class for all IceCube specific detector signal yield
    implementation methods for a point-like source. All IceCube detector signal
    yield implementation methods require a sinDec binning definition for
    the effective area. By default it is taken from the binning definitios
    stored in the dataset, but a user-defined sinDec binning can be specified
    if needed.
    """

    def __init__(self, sin_dec_binning=None, **kwargs):
        """Initializes a new detector signal yield implementation method
        object.

        Parameters
        ----------
        sin_dec_binning : BinningDefinition | None
            The BinningDefinition instance defining the sin(dec) binning that
            should be used to compute the sin(dec) dependency of the detector
            effective area. If set to None, the binning will be taken from the
            Dataset binning definitions.
        """
        super(PointLikeSourceI3DetSigYieldImplMethod, self).__init__(
            sin_dec_binning, **kwargs)

        # Define the supported source models.
        self.supported_sourcemodels = (PointLikeSource,)

    def source_to_array(self, sources):
        """Converts the sequence of PointLikeSource sources into a numpy record
        array holding the spatial information of the sources needed for the
        detector signal yield calculation.

        Parameters
        ----------
        sources : SourceModel | sequence of SourceModel
            The source model containing the spatial information of the source.

        Returns
        -------
        arr : numpy record ndarray
            The generated numpy record ndarray holding the spatial information
            for each source.
        """
        if(isinstance(sources, PointLikeSource)):
            sources = [ sources ]
        if(not issequenceof(sources, PointLikeSource)):
            raise TypeError('The source argument must be an instance of PointLikeSource!')

        arr = np.empty((len(sources),), dtype=[('dec', np.float64)])
        for (i, src) in enumerate(sources):
            arr['dec'][i] = src.dec

        return arr


class FixedFluxPointLikeSourceI3DetSigYield(I3DetSigYield):
    """The detector signal yield class for the
    FixedFluxPointLikeSourceI3DetSigYieldImplMethod detector signal yield
    implementation method.
    """
    def __init__(self, implmethod, dataset, fluxmodel, livetime, sin_dec_binning, log_spl_sinDec):
        """Constructs an IceCube detector signal yield instance for a
        point-like source with a fixed flux.

        Parameters
        ----------
        implmethod : FixedFluxPointLikeSourceI3DetSigYieldImplMethod instance
            The instance of the detector signal yield implementation
            method.
        dataset : Dataset instance
            The instance of Dataset holding the monte-carlo data this detector
            signal yield is made for.
        fluxmodel : FluxModel instance
            The instance of FluxModel with fixed parameters this detector signal
            yield is made for.
        livetime : float | Livetime instance
            The livetime in days or an instance of Livetime.
        sin_dec_binning : BinningDefinition instance
            The binning definition for sin(dec).
        log_spl_sinDec : scipy.interpolate.InterpolatedUnivariateSpline
            The spline instance representing the log value of the detector
            signal yield as a function of sin(dec).
        """
        if(not isinstance(implmethod, FixedFluxPointLikeSourceI3DetSigYieldImplMethod)):
            raise TypeError('The implmethod argument must be an instance of '
                'FixedFluxPointLikeSourceI3DetSigYieldImplMethod!')

        super(FixedFluxPointLikeSourceI3DetSigYield, self).__init__(
            implmethod, dataset, fluxmodel, livetime, sin_dec_binning)

        self.log_spl_sinDec = log_spl_sinDec

    @property
    def log_spl_sinDec(self):
        """The :class:`scipy.interpolate.InterpolatedUnivariateSpline` instance
        representing the spline for the log value of the detector signal
        yield as a function of sin(dec).
        """
        return self._log_spl_sinDec
    @log_spl_sinDec.setter
    def log_spl_sinDec(self, spl):
        if(not isinstance(spl, scipy.interpolate.InterpolatedUnivariateSpline)):
            raise TypeError('The log_spl_sinDec property must be an instance '
                'of scipy.interpolate.InterpolatedUnivariateSpline!')
        self._log_spl_sinDec = spl

    def __call__(self, src, src_flux_params=None):
        """Retrieves the detector signal yield for the list of given sources.

        Parameters
        ----------
        src : numpy record ndarray
            The numpy record ndarray with the field ``dec`` holding the
            declination of the source.
        src_flux_params : None
            Unused interface argument, because this implementation does not
            depend on any source flux fit parameters.

        Returns
        -------
        values : numpy 1d ndarray
            The array with the detector signal yield for each source.
        grads : None
            Because with this implementation the detector signal yield
            does not depend on any fit parameters. So there are no gradients
            and None is returned.
        """
        src_dec = np.atleast_1d(src['dec'])

        # Create results array.
        values = np.zeros_like(src_dec, dtype=np.float64)

        # Create mask for all source declinations which are inside the
        # declination range.
        mask = (np.sin(src_dec) >= self._sin_dec_binning.lower_edge)\
              &(np.sin(src_dec) <= self._sin_dec_binning.upper_edge)

        values[mask] = np.exp(self._log_spl_sinDec(np.sin(src_dec[mask])))

        return (values, None)


class FixedFluxPointLikeSourceI3DetSigYieldImplMethod(
    PointLikeSourceI3DetSigYieldImplMethod):
    """This detector signal yield implementation method constructs a
    detector signal yield for a fixed flux model, assuming a point-like
    source. This means that the detector signal yield does not depend on
    any source flux parameters, hence it is only dependent on the detector
    effective area.
    It constructs a one-dimensional spline function in sin(dec), using a
    :class:`scipy.interpolate.InterpolatedUnivariateSpline`.

    This detector signal yield implementation method works with all flux
    models.

    It is tailored to the IceCube detector at the South Pole, where the
    effective area depends soley on the zenith angle, and hence on the
    declination, of the source.
    """
    def __init__(self, sin_dec_binning=None, spline_order_sinDec=2, **kwargs):
        """Creates a new IceCube detector signal yield implementation
        method object for a fixed flux model. It requires a sinDec binning
        definition to compute the sin(dec) dependency of the detector effective
        area. The construct class method of this implementation method will
        create a spline function of a given order in logarithmic space of the
        effective area.

        Parameters
        ----------
        sin_dec_binning : BinningDefinition | None
            The BinningDefinition instance which defines the sin(dec) binning.
            If set to None, the binning will be taken from the Dataset binning
            definitions.
        spline_order_sinDec : int
            The order of the spline function for the logarithmic values of the
            detector signal yield along the sin(dec) axis.
            The default is 2.
        """
        super(FixedFluxPointLikeSourceI3DetSigYieldImplMethod, self).__init__(
            sin_dec_binning, **kwargs)

        self.supported_fluxmodels = (FluxModel,)

        self.spline_order_sinDec = spline_order_sinDec

    @property
    def spline_order_sinDec(self):
        """The order (int) of the logarithmic spline function, that splines the
        detector signal yield, along the sin(dec) axis.
        """
        return self._spline_order_sinDec
    @spline_order_sinDec.setter
    def spline_order_sinDec(self, order):
        if(not isinstance(order, int)):
            raise TypeError('The spline_order_sinDec property must be of '
                'type int!')
        self._spline_order_sinDec = order

    def construct_detsigyield(self, dataset, data, fluxmodel, livetime, ppbar=None):
        """Constructs a detector signal yield log spline function for the
        given fixed flux model.

        Parameters
        ----------
        dataset : Dataset instance
            The Dataset instance holding meta information about the data.
        data : DatasetData instance
            The DatasetData instance holding the monte-carlo event data.
            The numpy record ndarray holding the monte-carlo event data must
            contain the following data fields:

            - 'true_dec' : float
                The true declination of the data event.
            - 'true_energy' : float
                The true energy value of the data event.
            - 'mcweight' : float
                The monte-carlo weight of the data event in the unit
                GeV cm^2 sr.

        fluxmodel : FluxModel
            The flux model instance. Must be an instance of FluxModel.
        livetime : float | Livetime
            The live-time in days to use for the detector signal yield.
        ppbar : ProgressBar instance | None
            The instance of ProgressBar of the optional parent progress bar.

        Returns
        -------
        detsigyield : FixedFluxPointLikeSourceI3DetSigYield instance
            The DetSigYield instance for point-like source with a fixed flux.
        """
        # Check data types of the input arguments.
        super(FixedFluxPointLikeSourceI3DetSigYieldImplMethod, self).construct_detsigyield(
            dataset, data, fluxmodel, livetime)

        # Get integrated live-time in days.
        livetime_days = get_integrated_livetime_in_days(livetime)

        # Get the sin(dec) binning definition either as setting from this
        # implementation method, or from the dataset.
        sin_dec_binning = self.get_sin_dec_binning(dataset)

        # Calculate conversion factor from the flux model unit into the internal
        # flux unit GeV^-1 cm^-2 s^-1.
        toGeVcm2s = get_conversion_factor_to_internal_flux_unit(fluxmodel)

        # Calculate the detector signal yield contribution of each event.
        # The unit of mcweight is assumed to be GeV cm^2 sr.
        w = data.mc["mcweight"] * fluxmodel(data.mc["true_energy"])*toGeVcm2s * livetime_days * 86400.

        # Create a histogram along sin(true_dec).
        (h, bins) = np.histogram(np.sin(data.mc["true_dec"]),
                                 weights = w,
                                 bins = sin_dec_binning.binedges,
                                 density = False)

        # Normalize by solid angle of each bin which is
        # 2*\pi*(\Delta sin(\delta)).
        h /= (2.*np.pi * np.diff(sin_dec_binning.binedges))

        # Create spline in ln(h) at the histogram's bin centers.
        log_spl_sinDec = scipy.interpolate.InterpolatedUnivariateSpline(
            sin_dec_binning.bincenters, np.log(h), k=self.spline_order_sinDec)

        detsigyield = FixedFluxPointLikeSourceI3DetSigYield(
            self, dataset, fluxmodel, livetime, sin_dec_binning, log_spl_sinDec)

        return detsigyield


class PowerLawFluxPointLikeSourceI3DetSigYield(I3DetSigYield):
    """The detector signal yield class for the
    PowerLawFluxPointLikeSourceI3DetSigYieldImplMethod detector signal yield
    implementation method.
    """
    def __init__(self, implmethod, dataset, fluxmodel, livetime,
                 sin_dec_binning, log_spl_sinDec_gamma):
        """Constructs the detector signal yield instance.

        """
        if(not isinstance(implmethod, PowerLawFluxPointLikeSourceI3DetSigYieldImplMethod)):
            raise TypeError('The implmethod argument must be an instance of '
                'PowerLawFluxPointLikeSourceI3DetSigYieldImplMethod!')

        super(PowerLawFluxPointLikeSourceI3DetSigYield, self).__init__(
            implmethod, dataset, fluxmodel, livetime, sin_dec_binning)

        self.log_spl_sinDec_gamma = log_spl_sinDec_gamma

    @property
    def log_spl_sinDec_gamma(self):
        """The :class:`scipy.interpolate.RectBivariateSpline` instance
        representing the spline for the log value of the detector signal
        yield as a function of sin(dec) and gamma.
        """
        return self._log_spl_sinDec_gamma
    @log_spl_sinDec_gamma.setter
    def log_spl_sinDec_gamma(self, spl):
        if(not isinstance(spl, scipy.interpolate.RectBivariateSpline)):
            raise TypeError('The log_spl_sinDec_gamma property must be an '
                'instance of scipy.interpolate.RectBivariateSpline!')
        self._log_spl_sinDec_gamma = spl

    def __call__(self, src, src_flux_params):
        """Retrieves the detector signal yield for the given list of
        sources and their flux parameters.

        Parameters
        ----------
        src : numpy record ndarray
            The numpy record ndarray with the field ``dec`` holding the
            declination of the source.
        src_flux_params : (N_sources,)-shaped numpy record ndarray
            The numpy record ndarray containing the flux parameter ``gamma`` for
            the sources. ``gamma`` can be different for the different sources.

        Returns
        -------
        values : numpy (N_sources,)-shaped 1D ndarray
            The array with the detector signal yield for each source.
        grads : numpy (N_sources,N_fitparams)-shaped 2D ndarray
            The array containing the gradient values for each source and fit
            parameter. Since, this implementation depends on only one fit
            parameter, i.e. gamma, the array is (N_sources,1)-shaped.
        """
        src_dec = np.atleast_1d(src['dec'])
        src_gamma = src_flux_params['gamma']

        # Create results array.
        values = np.zeros_like(src_dec, dtype=np.float64)
        grads = np.zeros_like(src_dec, dtype=np.float64)

        # Calculate the detector signal yield only for the sources for
        # which we actually have detector acceptance. For the other sources,
        # the detector signal yield is zero.
        mask = (np.sin(src_dec) >= self._sin_dec_binning.lower_edge)\
              &(np.sin(src_dec) <= self._sin_dec_binning.upper_edge)

        if len(src_gamma) == len(src_dec):
            src_gamma = src_gamma[mask]
        else:
            src_gamma = src_gamma[0]

        values[mask] = np.exp(self._log_spl_sinDec_gamma(
            np.sin(src_dec[mask]), src_gamma, grid=False))
        grads[mask] = values[mask] * self._log_spl_sinDec_gamma(
            np.sin(src_dec[mask]), src_gamma, grid=False, dy=1)

        return (values, np.atleast_2d(grads))


class PowerLawFluxPointLikeSourceI3DetSigYieldImplMethod(
    PointLikeSourceI3DetSigYieldImplMethod, multiproc.IsParallelizable):
    """This detector signal yield implementation method constructs a
    detector signal yield for a variable power law flux model, which has
    the spectral index gamma as fit parameter, assuming a point-like source.
    It constructs a two-dimensional spline function in sin(dec) and gamma, using
    a :class:`scipy.interpolate.RectBivariateSpline`. Hence, the detector signal yield
    can vary with the declination and the spectral index, gamma, of the source.

    This detector signal yield implementation method works with a
    PowerLawFlux flux model.

    It is tailored to the IceCube detector at the South Pole, where the
    effective area depends soley on the zenith angle, and hence on the
    declination, of the source.
    """
    def __init__(
            self, gamma_grid, sin_dec_binning=None, spline_order_sinDec=2,
            spline_order_gamma=2, ncpu=None):
        """Creates a new IceCube detector signal yield implementation
        method object for a power law flux model. It requires a sinDec binning
        definition to compute the sin(dec) dependency of the detector effective
        area, and a gamma parameter grid to compute the gamma dependency of the
        detector signal yield.

        Parameters
        ----------
        gamma_grid : ParameterGrid instance
            The ParameterGrid instance which defines the grid of gamma values.
        sin_dec_binning : BinningDefinition | None
            The BinningDefinition instance which defines the sin(dec) binning.
            If set to None, the sin(dec) binning will be taken from the
            dataset's binning definitions.
        spline_order_sinDec : int
            The order of the spline function for the logarithmic values of the
            detector signal yield along the sin(dec) axis.
            The default is 2.
        spline_order_gamma : int
            The order of the spline function for the logarithmic values of the
            detector signal yield along the gamma axis.
            The default is 2.
        ncpu : int | None
            The number of CPUs to utilize. Global setting will take place if
            not specified, i.e. set to None.
        """
        super(PowerLawFluxPointLikeSourceI3DetSigYieldImplMethod, self).__init__(
            sin_dec_binning, ncpu=ncpu)

        self.supported_fluxmodels = (PowerLawFlux,)

        self.gamma_grid = gamma_grid
        self.spline_order_sinDec = spline_order_sinDec
        self.spline_order_gamma = spline_order_gamma

    @property
    def gamma_grid(self):
        """The ParameterGrid instance for the gamma grid that should be used for
        computing the gamma dependency of the detector signal yield.
        """
        return self._gamma_grid
    @gamma_grid.setter
    def gamma_grid(self, grid):
        if(not isinstance(grid, ParameterGrid)):
            raise TypeError('The gamma_grid property must be an instance of '
                'ParameterGrid!')
        self._gamma_grid = grid

    @property
    def spline_order_sinDec(self):
        """The order (int) of the logarithmic spline function, that splines the
        detector signal yield, along the sin(dec) axis.
        """
        return self._spline_order_sinDec
    @spline_order_sinDec.setter
    def spline_order_sinDec(self, order):
        if(not isinstance(order, int)):
            raise TypeError('The spline_order_sinDec property must be of '
                'type int!')
        self._spline_order_sinDec = order

    @property
    def spline_order_gamma(self):
        """The order (int) of the logarithmic spline function, that splines the
        detector signal yield, along the gamma axis.
        """
        return self._spline_order_gamma
    @spline_order_gamma.setter
    def spline_order_gamma(self, order):
        if(not isinstance(order, int)):
            raise TypeError('The spline_order_gamma property must be of '
                'type int!')
        self._spline_order_gamma = order

    def _get_signal_fitparam_names(self):
        """The list of signal fit parameter names the detector signal yield
        depends on.
        """
        return ['gamma']

    def construct_detsigyield(
            self, dataset, data, fluxmodel, livetime, ppbar=None):
        """Constructs a detector signal yield 2-dimensional log spline
        function for the given power law flux model with varying gamma values.

        Parameters
        ----------
        dataset : Dataset instance
            The Dataset instance holding the sin(dec) binning definition.
        data : DatasetData instance
            The DatasetData instance holding the monte-carlo event data.
            The numpy record array for the monte-carlo data of the dataset must
            contain the following data fields:

            - 'true_dec' : float
                The true declination of the data event.
            - 'mcweight' : float
                The monte-carlo weight of the data event in the unit
                GeV cm^2 sr.
            - 'true_energy' : float
                The true energy value of the data event.

        fluxmodel : FluxModel
            The flux model instance. Must be an instance of FluxModel.
        livetime : float | Livetime instance
            The live-time in days or an instance of Livetime to use for the
            detector signal yield.
        ppbar : ProgressBar instance | None
            The instance of ProgressBar of the optional parent progress bar.

        Returns
        -------
        detsigyield : PowerLawFluxPointLikeSourceI3DetSigYield instance
            The DetSigYield instance for a point-like source with a power law
            flux with variable gamma parameter.
        """
        # Check for the correct data types of the input arguments.
        super(PowerLawFluxPointLikeSourceI3DetSigYieldImplMethod, self).construct_detsigyield(
            dataset, data, fluxmodel, livetime)

        # Get integrated live-time in days.
        livetime_days = get_integrated_livetime_in_days(livetime)

        # Get the sin(dec) binning definition either as setting from this
        # implementation method, or from the dataset.
        sin_dec_binning = self.get_sin_dec_binning(dataset)

        # Calculate conversion factor from the flux model unit into the internal
        # flux unit GeV^-1 cm^-2 s^-1.
        toGeVcm2s = get_conversion_factor_to_internal_flux_unit(fluxmodel)

        # Define a function that creates a detector signal yield histogram
        # along sin(dec) for a given flux model, i.e. for given spectral index,
        # gamma.
        def hist(data_sin_true_dec, data_true_energy, sin_dec_binning, weights, fluxmodel):
            """Creates a histogram of the detector signal yield with the
            given sin(dec) binning.

            Parameters
            ----------
            data_sin_true_dec : 1d ndarray
                The sin(true_dec) values of the monte-carlo events.
            data_true_energy : 1d ndarray
                The true energy of the monte-carlo events.
            sin_dec_binning : BinningDefinition
                The sin(dec) binning definition to use for the histogram.
            weights : 1d ndarray
                The weight factors of each monte-carlo event where only the
                flux value needs to be multiplied with in order to get the
                detector signal yield.
            fluxmodel : FluxModel
                The flux model to get the flux values from.

            Returns
            -------
            h : 1d ndarray
                The numpy array containing the histogram values.
            """
            (h, edges) = np.histogram(data_sin_true_dec,
                                      bins = sin_dec_binning.binedges,
                                      weights = weights * fluxmodel(data_true_energy),
                                      density = False)
            return h

        data_sin_true_dec = np.sin(data.mc["true_dec"])
        weights = data.mc["mcweight"] * toGeVcm2s * livetime_days * 86400.

        # Make a copy of the gamma grid and extend the grid by one bin on each
        # side.
        gamma_grid = self._gamma_grid.copy()
        gamma_grid.add_extra_lower_and_upper_bin()

        # Construct the arguments for the hist function to be used in the
        # multiproc.parallelize function.
        args_list = [ ((data_sin_true_dec,
                        data.mc['true_energy'],
                        sin_dec_binning,
                        weights,
                        fluxmodel.copy({'gamma':gamma})), {})
                     for gamma in gamma_grid.grid ]
        h = np.vstack(
            multiproc.parallelize(
                hist, args_list, self.ncpu, ppbar=ppbar)).T

        # Normalize by solid angle of each bin along the sin(dec) axis.
        # The solid angle is given by 2*\pi*(\Delta sin(\delta))
        h /= (2.*np.pi * np.diff(sin_dec_binning.binedges)).reshape(
            (sin_dec_binning.nbins,1))

        log_spl_sinDec_gamma = scipy.interpolate.RectBivariateSpline(
            sin_dec_binning.bincenters, gamma_grid.grid, np.log(h),
            kx = self.spline_order_sinDec, ky = self.spline_order_gamma, s = 0)

        detsigyield = PowerLawFluxPointLikeSourceI3DetSigYield(
            self, dataset, fluxmodel, livetime, sin_dec_binning, log_spl_sinDec_gamma)

        return detsigyield



# -*- coding: utf-8 -*-

import abc
import itertools
import numpy as np

from skyllh.core.py import (
    classname,
    float_cast,
    issequenceof,
    typename
)
from skyllh.core.parameters import (
    FitParameter,
    make_params_hash
)
from skyllh.core.interpolate import (
    GridManifoldInterpolationMethod,
    Parabola1DGridManifoldInterpolationMethod
)
from skyllh.core.pdf import (
    PDF,
    PDFSet,
    IsBackgroundPDF,
    IsSignalPDF,
    SpatialPDF
)
from skyllh.core.timing import TaskTimer


class PDFRatio(object, metaclass=abc.ABCMeta):
    """Abstract base class for a PDF ratio class. It defines the interface
    of a PDF ratio class.
    """

    def __init__(self, pdf_type, *args, **kwargs):
        """Constructor for a PDF ratio class.

        Parameters
        ----------
        pdf_type : type
            The Python type of the PDF object the PDF ratio is made for.
        """
        super(PDFRatio, self).__init__(*args, **kwargs)

        self._pdf_type = pdf_type

    @property
    def n_fitparams(self):
        """(read-only) The number of fit parameters the PDF ratio depends on.
        This is the sum of signal and background fit parameters. At the moment
        only signal fit parameters are supported, so this property is equivalent
        to the n_signal_fitparams property. But this might change in the future.
        """
        return self.n_signal_fitparams

    @property
    def fitparam_names(self):
        """(read-only) The list of fit parameter names this PDF ratio is a
        function of.
        This is the superset of signal and background fit parameter names.
        At the moment only signal fit parameters are supported, so this property
        is equivalent to the signal_fitparam_names property. But this might
        change in the future.
        """
        return self.signal_fitparam_names

    @property
    def n_signal_fitparams(self):
        """(read-only) The number of signal fit parameters the PDF ratio depends
        on.
        """
        return len(self._get_signal_fitparam_names())

    @property
    def signal_fitparam_names(self):
        """(read-only) The list of signal fit parameter names this PDF ratio is
        a function of.
        """
        return self._get_signal_fitparam_names()

    @property
    def pdf_type(self):
        """(read-only) The Python type of the PDF object for which the PDF
        ratio is made for.
        """
        return self._pdf_type

    def _get_signal_fitparam_names(self):
        """This method must be re-implemented by the derived class and needs to
        return the list of signal fit parameter names, this PDF ratio is a
        function of. If it returns an empty list, the PDF ratio is independent
        of any signal fit parameters.

        Returns
        -------
        list of str
            The list of the signal fit parameter names, this PDF ratio is a
            function of. By default this method returns an empty list indicating
            that the PDF ratio depends on no signal parameter.
        """
        return []

    @abc.abstractmethod
    def get_ratio(self, tdm, params=None, tl=None):
        """Retrieves the PDF ratio value for each given trial data event, given
        the given set of fit parameters. This method is called during the
        likelihood maximization process.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance holding the trial data events for
            which the PDF ratio values should get calculated.
        params : dict | None
            The dictionary with the parameter name-value pairs.
            It can be ``None``, if the PDF ratio does not depend on any
            parameters.
        tl : TimeLord instance | None
            The optional TimeLord instance that should be used to measure
            timing information.

        Returns
        -------
        ratios : (N_events,)-shaped 1d numpy ndarray of float
            The PDF ratio value for each trial event.
        """
        pass

    @abc.abstractmethod
    def get_gradient(self, tdm, params, fitparam_name):
        """Retrieves the PDF ratio gradient for the parameter ``fitparam_name``
        for each given trial event, given the given set of fit parameters.
        This method is called during the likelihood maximization process.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance holding the trial data events for
            which the PDF ratio values should get calculated.
        params : dict
            The dictionary with the parameter names and values.
        fitparam_name : str
            The name of the fit parameter for which the gradient should
            get calculated.

        Returns
        -------
        gradient : (N_events,)-shaped 1d numpy ndarray of float
            The PDF ratio gradient value for each trial event.
        """
        pass


class SingleSourcePDFRatioArrayArithmetic(object):
    """This class provides arithmetic methods for arrays of PDFRatio instances.
    It has methods to calculate the product of the ratio values for a given set
    of PDFRatio objects. This class assumes a single source.

    The rational is that in the calculation of the derivates of the
    log-likelihood-ratio function for a given fit parameter, the product of the
    PDF ratio values of the PDF ratio objects which do not depend on that fit
    parameter is needed.
    """
    def __init__(self, pdfratios, fitparams):
        """Constructs a PDFRatio array arithmetic object assuming a single
        source.

        Parameters
        ----------
        pdfratios : list of PDFRatio
            The list of PDFRatio instances.
        fitparams : list of FitParameter
            The list of fit parameters. The order must match the fit parameter
            order of the minimizer.
        """
        super(SingleSourcePDFRatioArrayArithmetic, self).__init__()

        self.pdfratio_list = pdfratios
        self.fitparam_list = fitparams

        # The ``_ratio_values`` member variable will hold a
        # (N_pdfratios,N_events)-shaped array holding the PDF ratio values of
        # each PDF ratio object for each event. It will be created by the
        # ``initialize_for_new_trial`` method.
        self._ratio_values = None

        # Create a mapping of fit parameter index to pdfratio index. We
        # initialize the mapping with -1 first in order to be able to check in
        # the end if all fit parameters found a PDF ratio object.
        self._fitparam_idx_2_pdfratio_idx = np.repeat(
            np.array([-1], dtype=np.int64), len(self._fitparam_list))
        for ((fpidx, fitparam), (pridx, pdfratio)) in itertools.product(
                enumerate(self._fitparam_list), enumerate(self.pdfratio_list)):
            if(fitparam.name in pdfratio.fitparam_names):
                self._fitparam_idx_2_pdfratio_idx[fpidx] = pridx
        check_mask = (self._fitparam_idx_2_pdfratio_idx == -1)
        if(np.any(check_mask)):
            raise KeyError('%d fit parameters are not defined in any of the '
                'PDF ratio instances!'%(np.sum(check_mask)))

        # Create the list of indices of the PDFRatio instances, which depend on
        # at least one fit parameter.
        self._var_pdfratio_indices = np.unique(self._fitparam_idx_2_pdfratio_idx)

    def _precompute_static_pdfratio_values(self, tdm):
        """Pre-compute the PDF ratio values for the PDF ratios that do not
        depend on any fit parameters.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager that holds the trial event data for
            which the PDF ratio values should get calculated.
        """
        for (i, pdfratio) in enumerate(self._pdfratio_list):
            if(pdfratio.n_fitparams == 0):
                # The PDFRatio does not depend on any fit parameters. So we
                # pre-calculate the PDF ratio values for all the events. Since
                # the get_ratio method of the PDFRatio class might return a 2D
                # (N_sources, N_events)-shaped array, and we assume a single
                # source, we need to reshape the array, which does not involve
                # any data copying.
                self._ratio_values[i] = np.reshape(
                    pdfratio.get_ratio(tdm), (tdm.n_selected_events,))

    @property
    def pdfratio_list(self):
        """The list of PDFRatio objects.
        """
        return self._pdfratio_list
    @pdfratio_list.setter
    def pdfratio_list(self, seq):
        if(not issequenceof(seq, PDFRatio)):
            raise TypeError('The pdfratio_list property must be a sequence of '
                'PDFRatio instances!')
        self._pdfratio_list = list(seq)

    @property
    def fitparam_list(self):
        """The list of FitParameter instances.
        """
        return self._fitparam_list
    @fitparam_list.setter
    def fitparam_list(self, seq):
        if(not issequenceof(seq, FitParameter)):
            raise TypeError('The fitparam_list property must be a sequence of '
                'FitParameter instances!')
        self._fitparam_list = list(seq)

    def initialize_for_new_trial(self, tdm):
        """Initializes the PDFRatio array arithmetic for a new trial. For a new
        trial the data events change, hence we need to recompute the PDF ratio
        values of the fit parameter independent PDFRatio instances.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager that holds the trial event data for
            that this PDFRatioArrayArithmetic instance should get initialized.
        """
        n_events_old = 0
        if(self._ratio_values is not None):
            n_events_old = self._ratio_values.shape[1]

        # If the amount of events have changed, we need a new array holding the
        # ratio values.
        if(n_events_old != tdm.n_selected_events):
            # Create a (N_pdfratios,N_events)-shaped array to hold the PDF ratio
            # values of each PDF ratio object for each event.
            self._ratio_values = np.empty(
                (len(self._pdfratio_list), tdm.n_selected_events),
                dtype=np.float64)

        self._precompute_static_pdfratio_values(tdm)

    def get_pdfratio(self, idx):
        """Returns the PDFRatio instance that corresponds to the given fit
        parameter index.

        Parameters
        ----------
        fitparam_idx : int
            The index of the fit parameter.

        Returns
        -------
        pdfratio : PDFRatio
            The PDFRatio instance which corresponds to the given fit parameter
            index.
        """
        return self._pdfratio_list[idx]

    def calculate_pdfratio_values(self, tdm, fitparams, tl=None):
        """Calculates the PDF ratio values for the PDF ratio objects which
        depend on fit parameters.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager that holds the trial event data for
            which the PDF ratio values should get calculated.
        fitparams : dict
            The dictionary with the fit parameter name-value pairs.
        tl : TimeLord instance | None
            The optional TimeLord instance that should be used to measure
            timing information.
        """
        for (i, _pdfratio_i) in enumerate(self._pdfratio_list):
            # Since the get_ratio method of the PDFRatio class might return a 2D
            # (N_sources, N_events)-shaped array, and we assume a single source,
            # we need to reshape the array, which does not involve any data
            # copying.
            self._ratio_values[i] = np.reshape(
                _pdfratio_i.get_ratio(tdm, fitparams, tl=tl),
                (tdm.n_selected_events,))

    def get_ratio_product(self, excluded_idx=None):
        """Calculates the product of the of the PDF ratio values of each event,
        but excludes the PDF ratio values that correspond to the given excluded
        fit parameter index. This is useful for calculating the derivates of
        the log-likelihood ratio function.

        Parameters
        ----------
        excluded_fitparam_idx : int | None
            The index of the fit parameter whose PDF ratio values should get
            excluded from the product. If None, the product over all PDF ratio
            values will be computed.

        Returns
        -------
        product : 1D (N_events,)-shaped ndarray
            The product of the PDF ratio values for each event.
        """
        if(excluded_idx is None):
            return np.prod(self._ratio_values, axis=0)

        # Get the index of the PDF ratio object that corresponds to the excluded
        # fit parameter.
        #excluded_pdfratio_idx = self._fitparam_idx_2_pdfratio_idx[excluded_fitparam_idx]
        pdfratio_indices = list(range(self._ratio_values.shape[0]))
        pdfratio_indices.pop(excluded_idx)
        return np.prod(self._ratio_values[pdfratio_indices], axis=0)


class PDFRatioFillMethod(object, metaclass=abc.ABCMeta):
    """Abstract base class to implement a PDF ratio fill method. It can happen,
    that there are empty background bins but where signal could possibly be.
    A PDFRatioFillMethod implements what happens in such cases.
    """

    def __init__(self, *args, **kwargs):
        super(PDFRatioFillMethod, self).__init__(*args, **kwargs)

    @abc.abstractmethod
    def fill_ratios(self, ratios, sig_prob_h, bkg_prob_h,
                    sig_mask_mc_covered, sig_mask_mc_covered_zero_physics,
                    bkg_mask_mc_covered, bkg_mask_mc_covered_zero_physics):
        """The fill_ratios method is supposed to fill the ratio bins (array)
        with the signal / background division values. For bins (array elements),
        where the division is undefined, e.g. due to zero background, the fill
        method decides how to fill those bins.

        Note: Bins which have neither signal monte-carlo nor background
              monte-carlo coverage, are undefined about their signal-ness or
              background-ness by construction.

        Parameters
        ----------
        ratios : ndarray of float
            The multi-dimensional array for the final ratio bins. The shape is
            the same as the sig_h and bkg_h ndarrays.
        sig_prob_h : ndarray of float
            The multi-dimensional array (histogram) holding the signal
            probabilities.
        bkg_prob_h : ndarray of float
            The multi-dimensional array (histogram) holding the background
            probabilities.
        sig_mask_mc_covered : ndarray of bool
            The mask array indicating which array elements of sig_prob_h have
            monte-carlo coverage.
        sig_mask_mc_covered_zero_physics : ndarray of bool
            The mask array indicating which array elements of sig_prob_h have
            monte-carlo coverage but don't have physics contribution.
        bkg_mask_mc_covered : ndarray of bool
            The mask array indicating which array elements of bkg_prob_h have
            monte-carlo coverage.
            In case of experimental data as background, this mask indicate where
            (experimental data) background is available.
        bkg_mask_mc_covered_zero_physics : ndarray of bool
            The mask array ndicating which array elements of bkg_prob_h have
            monte-carlo coverage but don't have physics contribution.
            In case of experimental data as background, this mask contains only
            False entries.

        Returns
        -------
        ratios : ndarray
            The array holding the final ratio values.
        """
        return ratios

class Skylab2SkylabPDFRatioFillMethod(PDFRatioFillMethod):
    """This PDF ratio fill method implements the exact same fill method as in
    the skylab2 software named "skylab". It exists just for comparsion and
    backward compatibility reasons. In general, it should not be used, because
    it does not distinguish between bins with MC converage and physics model
    contribution, and those with MC coverage and no physics model contribution!
    """
    def __init__(self):
        super(Skylab2SkylabPDFRatioFillMethod, self).__init__()
        self.signallike_percentile = 99.

    def fill_ratios(self, ratio, sig_prob_h, bkg_prob_h,
                    sig_mask_mc_covered, sig_mask_mc_covered_zero_physics,
                    bkg_mask_mc_covered, bkg_mask_mc_covered_zero_physics):
        """Fills the ratio array.
        """
        # Check if we have predicted background for the entire background MC
        # range.
        if(np.any(bkg_mask_mc_covered_zero_physics)):
            raise ValueError('Some of the background bins have MC coverage but no physics background prediction. I don\'t know what to do in this case!')

        sig_domain = sig_prob_h > 0
        bkg_domain = bkg_prob_h > 0

        ratio[sig_domain & bkg_domain] = sig_prob_h[sig_domain & bkg_domain] / bkg_prob_h[sig_domain & bkg_domain]

        ratio_value = np.percentile(ratio[ratio > 1.], self.signallike_percentile)
        np.copyto(ratio, ratio_value, where=sig_domain & ~bkg_domain)

        return ratio

class MostSignalLikePDFRatioFillMethod(PDFRatioFillMethod):
    """PDF ratio fill method to set the PDF ratio to the most signal like PDF
    ratio for bins, where there is signal MC coverage but no background (MC)
    coverage.
    """
    def __init__(self, signallike_percentile=99.):
        """Creates the PDF ratio fill method object for filling PDF ratio bins,
        where there is signal MC coverage but no background (MC) coverage
        with the most signal-like ratio value.

        Parameters
        ----------
        signallike_percentile : float in range [0., 100.], default 99.
            The percentile of signal-like ratios, which should be taken as the
            ratio value for ratios with no background probability.
        """
        super(MostSignalLikePDFRatioFillMethod, self).__init__()

        self.signallike_percentile = signallike_percentile

    @property
    def signallike_percentile(self):
        """The percentile of signal-like ratios, which should be taken as the
        ratio value for ratios with no background probability. This percentile
        must be given as a float value in the range [0, 100] inclusively.
        """
        return self._signallike_percentile
    @signallike_percentile.setter
    def signallike_percentile(self, value):
        if(not isinstance(value, float)):
            raise TypeError('The signallike_percentile property must be of type float!')
        if(value < 0. or value > 100.):
            raise ValueError('The value of the signallike_percentile property must be in the range [0, 100]!')
        self._signallike_percentile = value

    def fill_ratios(self, ratio, sig_prob_h, bkg_prob_h,
                    sig_mask_mc_covered, sig_mask_mc_covered_zero_physics,
                    bkg_mask_mc_covered, bkg_mask_mc_covered_zero_physics):
        """Fills the ratio array.
        """
        # Check if we have predicted background for the entire background MC
        # range.
        if(np.any(bkg_mask_mc_covered_zero_physics)):
            raise ValueError('Some of the background bins have MC coverage but no physics background prediction. I don\'t know what to do in this case!')

        # Fill the bins where we have signal and background MC coverage.
        mask_sig_and_bkg_mc_covered = sig_mask_mc_covered & bkg_mask_mc_covered
        ratio[mask_sig_and_bkg_mc_covered] = sig_prob_h[mask_sig_and_bkg_mc_covered] / bkg_prob_h[mask_sig_and_bkg_mc_covered]

        # Calculate the ratio value, which should be used for ratio bins, where
        # we have signal MC coverage but no background MC coverage.
        ratio_value = np.percentile(ratio[ratio > 1.], self.signallike_percentile)
        mask_sig_but_notbkg_mc_covered = sig_mask_mc_covered & ~bkg_mask_mc_covered
        np.copyto(ratio, ratio_value, where=mask_sig_but_notbkg_mc_covered)

        return ratio


class MinBackgroundLikePDFRatioFillMethod(PDFRatioFillMethod):
    """PDF ratio fill method to set the PDF ratio to the minimal background like
    value for bins, where there is signal MC coverage but no background (MC)
    coverage.
    """
    def __init__(self):
        """Creates the PDF ratio fill method object for filling PDF ratio bins,
        where there is signal MC coverage but no background (MC) coverage
        with the minimal background-like ratio value.
        """
        super(MinBackgroundLikePDFRatioFillMethod, self).__init__()

    def fill_ratios(self, ratio, sig_prob_h, bkg_prob_h,
                    sig_mask_mc_covered, sig_mask_mc_covered_zero_physics,
                    bkg_mask_mc_covered, bkg_mask_mc_covered_zero_physics):
        """Fills the ratio array.
        """
        # Check if we have predicted background for the entire background MC
        # range.
        if(np.any(bkg_mask_mc_covered_zero_physics)):
            raise ValueError('Some of the background bins have MC coverage but no physics background prediction. I don\'t know what to do in this case!')

        # Fill the bins where we have signal and background MC coverage.
        mask_sig_and_bkg_mc_covered = sig_mask_mc_covered & bkg_mask_mc_covered
        ratio[mask_sig_and_bkg_mc_covered] = sig_prob_h[mask_sig_and_bkg_mc_covered] / bkg_prob_h[mask_sig_and_bkg_mc_covered]

        # Calculate the minimal background-like value.
        min_bkg_prob = np.min(bkg_prob_h[bkg_mask_mc_covered])

        # Set the ratio using the minimal background probability where we
        # have signal MC coverage but no background (MC) coverage.
        mask_sig_but_notbkg_mc_covered = sig_mask_mc_covered & ~bkg_mask_mc_covered
        ratio[mask_sig_but_notbkg_mc_covered] = sig_prob_h[mask_sig_but_notbkg_mc_covered] / min_bkg_prob

        return ratio


class SigOverBkgPDFRatio(PDFRatio):
    """This class implements a generic signal-over-background PDF ratio for a
    signal and a background PDF instance.
    It takes a signal PDF of type *pdf_type* and a background PDF of type
    *pdf_type* and calculates the PDF ratio.
    """
    def __init__(self, sig_pdf, bkg_pdf, pdf_type=None, same_axes=True,
        zero_bkg_ratio_value=1., *args, **kwargs):
        """Creates a new signal-over-background PDF ratio instance.

        Parameters
        ----------
        sig_pdf : class instance derived from `pdf_type`, IsSignalPDF
            The instance of the signal PDF.
        bkg_pdf : class instance derived from `pdf_type`, IsBackgroundPDF
            The instance of the background PDF.
        pdf_type : type | None
            The python type of the PDF object for which the PDF ratio is for.
            If set to None, the default class ``PDF`` will be used.
        same_axes : bool
            Flag if the signal and background PDFs are supposed to have the
            same axes. Default is True.
        zero_bkg_ratio_value : float
            The value of the PDF ratio to take when the background PDF value
            is zero. This is to avoid division by zero. Default is 1.
        """
        if(pdf_type is None):
            pdf_type = PDF

        super(SigOverBkgPDFRatio, self).__init__(
            pdf_type=pdf_type, *args, **kwargs)

        self.sig_pdf = sig_pdf
        self.bkg_pdf = bkg_pdf

        # Check that the PDF axes ranges are the same for the signal and
        # background PDFs.
        if(same_axes and (not sig_pdf.axes.is_same_as(bkg_pdf.axes))):
            raise ValueError('The signal and background PDFs do not have the '
                'same axes.')

        self.zero_bkg_ratio_value = zero_bkg_ratio_value

        # Define cache member variables to calculate gradients efficiently.
        self._cache_trial_data_state_id = None
        self._cache_params_hash = None
        self._cache_sigprob = None
        self._cache_bkgprob = None
        self._cache_siggrads = None
        self._cache_bkggrads = None

    @property
    def sig_pdf(self):
        """The signal PDF object used to create the PDF ratio.
        """
        return self._sig_pdf
    @sig_pdf.setter
    def sig_pdf(self, pdf):
        if(not isinstance(pdf, self.pdf_type)):
            raise TypeError('The sig_pdf property must be an instance of '
                '%s!'%(typename(self.pdf_type)))
        if(not isinstance(pdf, IsSignalPDF)):
            raise TypeError('The sig_pdf property must be an instance of '
                'IsSignalPDF!')
        self._sig_pdf = pdf

    @property
    def bkg_pdf(self):
        """The background PDF object used to create the PDF ratio.
        """
        return self._bkg_pdf
    @bkg_pdf.setter
    def bkg_pdf(self, pdf):
        if(not isinstance(pdf, self.pdf_type)):
            raise TypeError('The bkg_pdf property must be an instance of '
                '%s!'%(typename(self.pdf_type)))
        if(not isinstance(pdf, IsBackgroundPDF)):
            raise TypeError('The bkg_pdf property must be an instance of '
                'IsBackgroundPDF!')
        self._bkg_pdf = pdf

    @property
    def zero_bkg_ratio_value(self):
        """The value of the PDF ratio to take when the background PDF value
        is zero. This is to avoid division by zero.
        """
        return self._zero_bkg_ratio_value
    @zero_bkg_ratio_value.setter
    def zero_bkg_ratio_value(self, v):
        v = float_cast(v, 'The zero_bkg_ratio_value must be castable into a '
            'float!')
        self._zero_bkg_ratio_value = v

    def _get_signal_fitparam_names(self):
        """Returns the list of fit parameter names the signal PDF depends on.
        """
        return self._sig_pdf.param_set.floating_param_name_list

    def get_ratio(self, tdm, params=None, tl=None):
        """Calculates the PDF ratio for the given trial events.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance holding the trial data events for
            which the PDF ratio values should be calculated.
        params : dict | None
            The dictionary holding the parameter names and values for which the
            probability ratio should get calculated.
            This can be ``None``, if the signal and background PDFs do not
            depend on any parameters.
        tl : TimeLord instance | None
            The optional TimeLord instance that should be used to measure
            timing information.

        Returns
        -------
        ratios : (N_events)-shaped numpy ndarray
            The ndarray holding the probability ratio for each event (and each
            source). The dimensionality of the returned ndarray depends on the
            dimensionality of the probability ndarray returned by the
            ``get_prob`` method of the signal PDF object.
        """
        with TaskTimer(tl, 'Get sig prob.'):
            (sigprob, self._cache_siggrads) = self._sig_pdf.get_prob(
                tdm, params, tl=tl)
        with TaskTimer(tl, 'Get bkg prob.'):
            (bkgprob, self._cache_bkggrads) = self._bkg_pdf.get_prob(
                tdm, params, tl=tl)

        with TaskTimer(tl, 'Calc PDF ratios.'):
            # Select only the events, where background pdf is greater than zero.
            m = (bkgprob > 0)
            ratios = np.full_like(sigprob, self._zero_bkg_ratio_value)
            ratios[m] = sigprob[m] / bkgprob[m]

        # Store the current state of parameter values and trial data, so that
        # the get_gradient method can verify the consistency of the signal and
        # background probabilities and gradients.
        self._cache_trial_data_state_id = tdm.trial_data_state_id
        self._cache_params_hash = make_params_hash(params)
        self._cache_sigprob = sigprob
        self._cache_bkgprob = bkgprob

        return ratios

    def get_gradient(self, tdm, params, fitparam_name):
        """Retrieves the gradient of the PDF ratio w.r.t. the given fit
        parameter. This method must be called after the ``get_ratio`` method.

        Parameters
        ----------
        tdm : TrialDataManager instance
            The instance of TrialDataManager that should be used to get the
            trial data from.
        params : dict
            The dictionary with the parameter names and values.
        fitparam_name : str
            The name of the fit parameter for which the gradient should
            get calculated.

        Returns
        -------
        gradient : (N_events,)-shaped 1d numpy ndarray of float
            The PDF ratio gradient value for each trial event.
        """
        if((tdm.trial_data_state_id != self._cache_trial_data_state_id) or
           (make_params_hash(params) != self._cache_params_hash)):
            raise RuntimeError('The get_ratio method must be called prior to '
                'the get_gradient method!')

        # Create the 1D return array for the gradient.
        grad = np.zeros((tdm.n_selected_events,), dtype=np.float64)

        # Calculate the gradient for the given fit parameter.
        # There are four cases:
        #   1) Neither the signal nor the background PDF depend on the fit
        #      parameter.
        #   2) Only the signal PDF depends on the fit parameter.
        #   3) Only the background PDF depends on the fit parameter.
        #   4) Both, the signal and the background PDF depend on the fit
        #      parameter.
        sig_pdf_param_set = self._sig_pdf.param_set
        bkg_pdf_param_set = self._bkg_pdf.param_set

        sig_dep = sig_pdf_param_set.has_floating_param(fitparam_name)
        bkg_dep = bkg_pdf_param_set.has_floating_param(fitparam_name)

        if(sig_dep and (not bkg_dep)):
            # Case 2, which should be the most common case.
            # Get the signal grad idx for that fit parameter.
            sig_pidx = sig_pdf_param_set.get_floating_pidx(fitparam_name)
            bkgprob = self._cache_bkgprob
            m = bkgprob > 0
            grad[m] = self._cache_siggrads[sig_pidx][m] / bkgprob[m]
            return grad
        if((not sig_dep) and (not bkg_dep)):
            # Case 1. Returns zeros.
            return grad

        if(sig_dep and bkg_dep):
            # Case 4.
            sig_pidx = sig_pdf_param_set.get_floating_pidx(fitparam_name)
            bkg_pidx = bkg_pdf_param_set.get_floating_pidx(fitparam_name)
            m = self._cache_bkgprob > 0
            s = self._cache_sigprob[m]
            b = self._cache_bkgprob[m]
            sgrad = self._cache_siggrads[sig_pidx][m]
            bgrad = self._cache_bkggrads[bkg_pidx][m]
            # Make use of quotient rule of differentiation.
            grad[m] = (sgrad * b - bgrad * s) / b**2
            return grad

        # Case 3.
        bkg_pidx = bkg_pdf_param_set.get_floating_pidx(fitparam_name)
        bkgprob = self._cache_bkgprob
        m = bkgprob > 0
        grad[m] = (-self._cache_sigprob[m] / bkgprob[m]**2 *
            self._cache_bkggrads[bkg_pidx][m])
        return grad


class SpatialSigOverBkgPDFRatio(SigOverBkgPDFRatio):
    """This class implements a signal-over-background PDF ratio for spatial
    PDFs. It takes a signal PDF of type SpatialPDF and a background PDF of type
    SpatialPDF and calculates the PDF ratio.
    """
    def __init__(self, sig_pdf, bkg_pdf, *args, **kwargs):
        """Creates a new signal-over-background PDF ratio instance for spatial
        PDFs.

        Parameters
        ----------
        sig_pdf : class instance derived from SpatialPDF, IsSignalPDF
            The instance of the spatial signal PDF.
        bkg_pdf : class instance derived from SpatialPDF, IsBackgroundPDF
            The instance of the spatial background PDF.
        """
        super(SpatialSigOverBkgPDFRatio, self).__init__(pdf_type=SpatialPDF,
            sig_pdf=sig_pdf, bkg_pdf=bkg_pdf, *args, **kwargs)

        # Make sure that the PDFs have two dimensions, i.e. RA and Dec.
        if(not sig_pdf.ndim == 2):
            raise ValueError('The spatial signal PDF must have two dimensions! '
                'Currently it has %d!'%(sig_pdf.ndim))


class SigSetOverBkgPDFRatio(PDFRatio):
    """Class for a PDF ratio class that takes a PDFSet of PDF type
    *pdf_type* as signal PDF and a PDF of type *pdf_type* as background PDF.
    The signal PDF depends on signal fit parameters and a interpolation method
    defines how the PDF ratio gets interpolated between the fit parameter
    values.
    """
    def __init__(self, pdf_type, signalpdfset, backgroundpdf,
                 interpolmethod=None, *args, **kwargs):
        """Constructor called by creating an instance of a class which is
        derived from this PDFRatio class.

        Parameters
        ----------
        pdf_type : type
            The Python type of the PDF object for which the PDF ratio is for.
        signalpdfset : class instance derived from PDFSet (for PDF type
                       ``pdf_type``), and IsSignalPDF
            The PDF set, which provides signal PDFs for a set of
            discrete signal fit parameters.
        backgroundpdf : class instance derived from ``pdf_type``, and
                        IsBackgroundPDF
            The background PDF instance.
        interpolmethod : class of GridManifoldInterpolationMethod | None
            The class implementing the fit parameter interpolation method for
            the PDF ratio manifold grid.
            If set to None (default), the
            Parabola1DGridManifoldInterpolationMethod will be used for
            1-dimensional parameter manifolds.
        """
        # Call super to allow for multiple class inheritance.
        super(SigSetOverBkgPDFRatio, self).__init__(pdf_type, *args, **kwargs)

        self.signalpdfset = signalpdfset
        self.backgroundpdf = backgroundpdf

        # Define the default fit parameter interpolation method. The default
        # depends on the dimensionality of the fit parameter manifold.
        if(interpolmethod is None):
            ndim = signalpdfset.fitparams_grid_set.ndim
            if(ndim == 1):
                interpolmethod = Parabola1DGridManifoldInterpolationMethod
            else:
                raise ValueError('There is no default fit parameter manifold grid interpolation method available for %d dimensions!'%(ndim))
        self.interpolmethod = interpolmethod

        # Generate the list of signal fit parameter names once here.
        self._cache_signal_fitparam_name_list = self.signal_fitparam_names

    @property
    def backgroundpdf(self):
        """The background PDF object, derived from ``pdf_type`` and
        IsBackgroundPDF.
        """
        return self._bkgpdf
    @backgroundpdf.setter
    def backgroundpdf(self, pdf):
        if(not (isinstance(pdf, self.pdf_type) and isinstance(pdf, IsBackgroundPDF))):
            raise TypeError('The backgroundpdf property must be an object which is derived from %s and IsBackgroundPDF!'%(typename(self.pdf_type)))
        self._bkgpdf = pdf

    @property
    def signalpdfset(self):
        """The signal PDFSet object for ``pdf_type`` PDF objects.
        """
        return self._sigpdfset
    @signalpdfset.setter
    def signalpdfset(self, pdfset):
        if(not (isinstance(pdfset, PDFSet) and isinstance(pdfset, IsSignalPDF) and issubclass(pdfset.pdf_type, self.pdf_type))):
            raise TypeError('The signalpdfset property must be an object which is derived from PDFSet and IsSignalPDF and whose pdf_type property is a subclass of %s!'%(typename(self.pdf_type)))
        self._sigpdfset = pdfset

    @property
    def interpolmethod(self):
        """The class derived from GridManifoldInterpolationMethod
        implementing the interpolation of the fit parameter manifold.
        """
        return self._interpolmethod
    @interpolmethod.setter
    def interpolmethod(self, cls):
        if(not issubclass(cls, GridManifoldInterpolationMethod)):
            raise TypeError('The interpolmethod property must be a sub-class '
                'of GridManifoldInterpolationMethod!')
        self._interpolmethod = cls

    def _get_signal_fitparam_names(self):
        """Returns the list of signal fit parameter names this PDF ratio is a
        function of. The list is taken from the fit parameter grid set of the
        signal PDFSet object. By construction this parameter grid set defines
        the signal fit parameters.
        """
        return self._sigpdfset.fitparams_grid_set.parameter_names

    def convert_signal_fitparam_name_into_index(self, signal_fitparam_name):
        """Converts the given signal fit parameter name into the parameter
        index, i.e. the position of parameter in the signal parameter grid set.

        Parameters
        ----------
        signal_fitparam_name : str
            The name of the signal fit parameter.

        Returns
        -------
        index : int
            The index of the signal fit parameter.
        """
        # If there is only one signal fit parameter, we just return index 0.
        if(len(self._cache_signal_fitparam_name_list) == 1):
            return 0

        # At this point we have to loop through the list and do name
        # comparisons.
        for (index, name) in enumerate(self._cache_signal_fitparam_name_list):
            if(name == signal_fitparam_name):
                return index

        # At this point there is no parameter defined.
        raise KeyError('The PDF ratio "%s" has no signal fit parameter named '
            '"%s"!'%(classname(self), signal_fitparam_name))



# -*- coding: utf-8 -*-

import unittest
import numpy as np

from skyllh.i3.coords import (
    azi_to_ra_transform,
    ra_to_azi_transform,
    hor_to_equ_transform
)


class TestCoords(unittest.TestCase):
    """Test IceCube specific coordinate utility functions.
    """
    def setUp(self):
        self.mjd = 58457
        self.azi = 0.5
        self.zen = 0.5
        self.ra = 2.356062973186092
        self.dec = 2.641592653589793

    def test_azi_to_ra_transform(self):
        self.assertEqual(azi_to_ra_transform(self.azi, self.mjd), self.ra)

    def test_ra_to_azi_transform(self):
        self.assertEqual(ra_to_azi_transform(self.ra, self.mjd), self.azi)

    def test_hor_to_equ_transform(self):
        self.assertEqual(hor_to_equ_transform(self.azi, self.zen, self.mjd),
                         (self.ra, self.dec))


if(__name__ == '__main__'):
    unittest.main()
